"""Genrate date for the frontdoor graph."""

import numpy as np
import pandas as pd

from y0.dsl import Variable, W, X, Y, Z

__all__ = [
    "generate_data_for_frontdoor_backdoor",
]


def generate_data_for_frontdoor_backdoor(
    num_samples: int,
    treatments: dict[Variable, float] | None = None,
    *,
    seed: int | None = None,
) -> pd.DataFrame:
    """Generate testing data for the front-door back_door case study.

    :param num_samples: The number of samples to generate. Try 1000.
    :param treatments: An optional dictionary of the values to fix each variable to.
    :param seed: An optional random seed for reproducibility purposes
    :returns: A pandas Dataframe with columns corresponding
        to the variable names in the front-door back_door example
    """
    if treatments is None:
        treatments = {}
    generator = np.random.default_rng(seed)
    if W in treatments:
        w = np.full(num_samples, treatments[W])
    else:
        w = generator.binomial(n=1, p=0.3, size=num_samples)

    beta0_x = 0.1
    beta_w_to_x = 0.5
    if X in treatments:
        x = np.full(num_samples, treatments[X])
    else:
        p_x = beta0_x + w * beta_w_to_x
        x = generator.binomial(n=1, p=p_x, size=num_samples)

    beta0_z = 0.5
    beta_x_to_z = 0.4
    if Z in treatments:
        z = np.full(num_samples, treatments[Z])
    else:
        p_z = beta0_z + x * beta_x_to_z
        z = generator.binomial(n=1, p=p_z, size=num_samples)

    beta0_y = 0.1
    beta_z_to_y = 0.7
    beta_w_to_y = 0.1
    if Y in treatments:
        y = np.full(num_samples, treatments[Y])
    else:
        p_y = beta0_y + z * beta_z_to_y + w * beta_w_to_y
        y = generator.binomial(n=1, p=p_y, size=num_samples)

    return pd.DataFrame({X.name: x, Z.name: z, W.name: w, Y.name: y})



# -*- coding: utf-8 -*-

"""Generate data using a linear structural causal model."""

from __future__ import annotations

import itertools as itt
from functools import partial
from typing import (
    Any,
    Callable,
    Dict,
    FrozenSet,
    Mapping,
    MutableMapping,
    NamedTuple,
    Optional,
    Tuple,
    Union,
    cast,
)

import pandas as pd
from numpy.random import normal, uniform
from sklearn.linear_model import LinearRegression
from tqdm.auto import trange

from .algorithm.conditional_independencies import get_conditional_independencies
from .dsl import V1, V2, V3, V4, V5, V6, Variable
from .graph import NxMixedGraph
from .struct import DSeparationJudgement

__all__ = [
    "LinearSCM",
    "simulate",
]

example_graph = NxMixedGraph.from_edges(
    directed=[
        (V1, V2),
        (V1, V4),
        (V2, V5),
        (V4, V5),
        (V4, V6),
        (V3, V5),
        (V5, V6),
    ],
)
example_generators = {
    V1: partial(uniform, low=-1.0, high=1.0),
    V2: partial(uniform, low=-2.0, high=2.0),
    V3: partial(normal, loc=0.0, scale=1.0),
    V4: partial(normal, loc=0.0, scale=2.0),
    V5: partial(uniform, low=-3.0, high=3.0),
    V6: partial(normal, loc=0.0, scale=3.0),
}


class FitTuple(NamedTuple):
    """A tuple representing the linear regression fit."""

    regression: LinearRegression
    slope: float
    intercept: float
    r2: float
    d_separation: Optional[DSeparationJudgement]


def simulate(
    graph: NxMixedGraph,
    trials: int = 200,
    return_fits: bool = True,
    tqdm_kwargs: Optional[Mapping[str, Any]] = None,
    **kwargs,
) -> Union[pd.DataFrame, Tuple[pd.DataFrame, Mapping[FrozenSet[Variable], FitTuple]]]:
    """Simulate a graph using a linear structural causal model."""
    judgements = get_conditional_independencies(graph)
    cis: Mapping[FrozenSet[Variable], DSeparationJudgement] = {
        frozenset((judgement.left, judgement.right)): judgement for judgement in judgements
    }

    _tqdm_kwargs = dict(desc="Simulation", unit="trial", unit_scale=True)
    if tqdm_kwargs:
        _tqdm_kwargs.update(tqdm_kwargs)

    linear_scm = LinearSCM(graph, **kwargs)
    results = {
        trial: {variable.name: values for variable, values in linear_scm.trial().items()}
        for trial in trange(trials, **_tqdm_kwargs)
    }
    rv = pd.DataFrame(results).T

    # Get a prioritized ordering for edges actually in graph
    order = list(graph.directed.edges())
    _x = {frozenset(edge) for edge in order}
    order.extend(
        (left, right)
        for left, right in itt.combinations(sorted(graph.nodes(), key=lambda n: n.name), 2)
        if frozenset((left, right)) not in _x
    )

    if not return_fits:
        return rv

    fits: Dict[FrozenSet[Variable], FitTuple] = {}
    for parent, child in order:
        x, y = rv[parent.name].to_numpy().reshape(-1, 1), rv[child.name]
        regression = LinearRegression()
        regression.fit(x, y)
        r2 = regression.score(x, y)
        key = frozenset((parent, child))
        fits[key] = FitTuple(
            regression=regression,
            slope=regression.coef_[0],
            intercept=regression.intercept_,
            r2=r2,
            d_separation=cis.get(key),
        )

    return rv, fits


Generator = Callable[[], float]


class LinearSCM:
    """A data structure for a simulation.

    Provides an implementation of the simulation presented in Figure 1 and Example 2.2 in
    `Graphical criteria for efficient total effect estimation via adjustment in causal
    linear models <https://arxiv.org/abs/1907.02435>`_
    """

    #: The generator functions for all nodes
    generators: Mapping[Variable, Generator]

    #: Weights corresponding to each edge
    weights: Mapping[Tuple[Variable, Variable], float]

    def __init__(
        self,
        graph: NxMixedGraph,
        generators: Optional[Mapping[Variable, Generator]] = None,
        weights: Optional[Mapping[Tuple[Variable, Variable], float]] = None,
    ) -> None:
        """Prepare a simulation.

        :param graph: The ADMG
        :param generators: Generator functions for each node. If none given, defaults to uniformly
            distributed between -1.0 and 1.0.
        :param weights: Weights for each directed edge. If none given, defaults to uniformly distributed
            weights between -1.0 and 1.0.
        :raises ValueError: if the generators or weights are given and don't match the graph
        """
        self.graph = graph

        if weights is None:
            self.weights = {edge: uniform(low=0.0, high=1.0) for edge in graph.directed.edges()}
        elif set(weights) != set(self.graph.directed.edges()):
            raise ValueError("given weights do not exactly match directed edges in the graph")
        else:
            self.weights = weights

        if generators is None:
            self.generators = cast(
                Mapping[Variable, Generator],
                {node: partial(uniform, low=-1.0, high=1.0) for node in self.graph.nodes()},
            )
        elif set(generators) != set(self.graph.nodes()):
            raise ValueError("given node generators do not exactly match nodes in the graph")
        else:
            self.generators = generators

    def generate(self, node: Variable) -> float:
        """Generate a value for the variable."""
        return self.generators[node]()

    def fix(self, values: Mapping[Variable, float]) -> LinearSCM:
        """Create a new simulation with the given nodes fixed to their values."""

        def _get(node: Variable) -> float:
            return values[node]

        generators = cast(
            Mapping[Variable, Generator],
            {
                node: generator if node not in values else partial(_get, node=node)
                for node, generator in self.generators.items()
            },
        )
        return LinearSCM(graph=self.graph, generators=generators, weights=self.weights)

    def trial(self) -> Mapping[Variable, float]:
        """Perform a single trial.

        :return: A mapping from variable names to simulation values
        """
        values: MutableMapping[Variable, float] = {}
        for node in self.graph.topological_sort():
            values[node] = self.generate(node) + sum(
                self.weights[parent, node] * values[parent]
                for parent, _ in self.graph.directed.in_edges(node)
            )
        return values



"""Test sigma separation."""

import unittest

from y0.algorithm.conditional_independencies import are_d_separated
from y0.algorithm.sigma_separation import (
    are_sigma_separated,
    get_equivalence_classes,
    is_collider,
    is_non_collider_fork,
    is_non_collider_left_chain,
    is_non_collider_right_chain,
    is_z_sigma_open,
)
from y0.dsl import V1, V2, V3, V4, V5, V6, Variable
from y0.graph import NxMixedGraph

V7, V8 = map(Variable, ["V7", "V8"])

#: Figure 3 from https://arxiv.org/abs/1807.03024
graph = NxMixedGraph.from_edges(
    directed=[
        (V1, V2),
        (V2, V3),
        (V3, V4),
        (V4, V5),
        (V4, V8),
        (V5, V2),
        (V6, V7),
        (V7, V6),
    ],
    undirected=[
        (V1, V2),
        (V4, V6),
        (V4, V7),
        (V6, V7),
    ],
)


class TestSigmaSeparation(unittest.TestCase):
    """Test sigma separation.

    These tests come from Table 1 in https://arxiv.org/abs/1807.03024.
    The sigma equivalence classes in Figure 3 are {v1}, {v2, v3, v4, v5},
    {v6, v7}, and {v8}.
    """

    def setUp(self) -> None:
        """Set up the test case."""
        self.sigma = get_equivalence_classes(graph)

    def test_equivalence_classes(self):
        """Test getting equivalence classes."""
        equivalent_classes = {
            frozenset([V1]),
            frozenset([V2, V3, V4, V5]),
            frozenset([V6, V7]),
            frozenset([V8]),
        }
        expected_equivalent_classes = {n: c for c in equivalent_classes for n in c}
        self.assertEqual(expected_equivalent_classes, self.sigma)

    def test_collider(self):
        """Test checking colliders."""
        self.assertTrue(is_collider(graph, left=V4, middle=V5, right=V4, conditions={V3, V5}))

    def test_left_chain(self):
        """Test checking non-colliders (left chain)."""
        self.assertTrue(
            is_non_collider_left_chain(
                graph, left=V5, middle=V4, right=V6, conditions={V3, V5}, sigma=self.sigma
            )
        )

    def test_right_chain(self):
        """Test checking non-colliders (right chain)."""
        self.assertTrue(
            is_non_collider_right_chain(
                graph, left=V1, middle=V2, right=V3, conditions={V3, V5}, sigma=self.sigma
            )
        )
        self.assertTrue(
            is_non_collider_right_chain(
                graph, left=V2, middle=V3, right=V4, conditions={V3, V5}, sigma=self.sigma
            )
        )
        self.assertTrue(
            is_non_collider_right_chain(
                graph, left=V3, middle=V4, right=V5, conditions={V3, V5}, sigma=self.sigma
            )
        )

    def test_fork(self):
        """Test checking non-colliders (fork)."""
        self.assertTrue(
            is_non_collider_fork(
                graph, left=V5, middle=V4, right=V8, conditions={V3, V5}, sigma=self.sigma
            )
        )

    def test_z_sigma_open(self):
        """Tests for z-sigma-open paths."""
        # this is a weird example since it backtracks
        path = [V1, V2, V3, V4, V5, V4, V6]
        self.assertFalse(is_z_sigma_open(graph, path, sigma=self.sigma))
        self.assertTrue(is_z_sigma_open(graph, path, conditions={V3, V5}, sigma=self.sigma))

    def test_separations_figure_3(self):
        """Test comparisons of d-separation and sigma-separation."""
        for left, right, conditions, d, s in [
            (V2, V4, [V3, V5], True, False),
            (V1, V6, [], True, True),
            (V1, V6, [V3, V5], True, False),
            (V1, V8, [], False, False),
            (V1, V8, [V3, V5], True, False),
            (V1, V8, [V4], True, True),
        ]:
            with self.subTest(left=left, right=right, conditions=conditions):
                self.assertEqual(
                    d, are_d_separated(graph, left, right, conditions=conditions).separated
                )
                self.assertEqual(s, are_sigma_separated(graph, left, right, conditions=conditions))



"""Implementation of sigma-separation."""

from typing import Iterable, Optional, Sequence

import networkx as nx
from more_itertools import triplewise

from y0.dsl import Variable
from y0.graph import NxMixedGraph

__all__ = [
    "are_sigma_separated",
    "is_z_sigma_open",
    "get_equivalence_classes",
]


def are_sigma_separated(
    graph: NxMixedGraph,
    left: Variable,
    right: Variable,
    *,
    conditions: Optional[Iterable[Variable]] = None,
    cutoff: Optional[int] = None,
) -> bool:
    """Test if two variables are sigma-separated.

    Sigma separation is a generalization of d-separation that
    works not only for directed acyclic graphs, but also for
    directed graphs containing cycles. It was originally introduced
    in https://arxiv.org/abs/1807.03024.

    We say that X and Y are σ-connected by Z or not
    σ-separated by Z if there exists a path π (with some
    n ≥ 1 nodes) in G with one endnode in X and
    one endnode in Y that is Z-σ-open. σ-separated is the
    opposite of σ-connected (logical not).

    :param graph: Graph to test
    :param left: A node in the graph
    :param right: A node in the graph
    :param conditions: A collection of graph nodes
    :param cutoff: The maximum path length to check. By default, is unbounded.
    :return: If a and b are sigma-separated.
    """
    if conditions is None:
        conditions = set()
    else:
        conditions = set(conditions)

    sigma = get_equivalence_classes(graph)
    return not any(
        is_z_sigma_open(graph, path, conditions=conditions, sigma=sigma)
        # Technically, this algorithm should generate all paths, which could include
        # repeat visits to nodes and edges, but this is computationally intractable,
        # so the is_z_sigma_open() subroutine contains a novel path augmentation
        # algorithm. This might not be officially complete.
        for path in nx.all_simple_paths(graph.disorient(), left, right, cutoff=cutoff)
    )


def is_z_sigma_open(
    graph: NxMixedGraph,
    path: Sequence[Variable],
    *,
    sigma: dict[Variable, set[Variable]],
    conditions: Optional[set[Variable]] = None,
) -> bool:
    r"""Check if a path is Z-sigma-open.

    :param graph: A mixed graph
    :param path: A path in the graph. Denoted as $\pi$ in the paper. The
        node in position $i$ in the path is denoted with $v_i$.
    :param conditions: A set of nodes chosen as conditions, denoted by $Z$ in the paper
    :param sigma: The set of equivalence classes. Can be calculated with
        :func:`get_equivalence_classes`, denoted by $\sigma(v)$ in the paper.
    :returns: If the path is Z-sigma-open

    A path is $Z-\sigma-\text{open}$ if:

    1. The end nodes $v_1, v_n \notin Z$
    2. Every triple of adjacent nodes in the path is of the form

       1. Collider (:func:`is_collider`)
       2. (non-collider) left chain (:func:`is_non_collider_left_chain`)
       3. (non-collider) right chain (:func:`is_non_collider_left_chain`)
       4. (non-collider) fork (:func:`is_non_collider_fork`)
       5. (non-collider) with undirected edge (:func:`is_non_collider_undirected`, not implemented)
    """
    if conditions is None:
        conditions = set()
    if path[0] in conditions or path[-1] in conditions:
        return False
    return all(
        _triple_has_correct_form(graph, left, middle, right, conditions, sigma)
        for left, middle, right in triplewise(path)
    )


def _triple_has_correct_form(
    graph: NxMixedGraph,
    left: Variable,
    middle: Variable,
    right: Variable,
    conditions: set[Variable],
    sigma: dict[Variable, set[Variable]],
) -> bool:
    if _triple_helper(graph, left, middle, right, conditions, sigma):
        return True
    # augment with backtracks, since you're allowed to go back (just like Season 5 of Lost).
    # this is a better solution than generating infinite paths, but might still be mathematically
    # incomplete. In this setup, 𝑣3→𝑣4↔𝑣6 becomes 𝑣3→𝑣4→𝑣5←𝑣4↔𝑣6 to get some sweet backtrack paths
    # through the middle node to a neighbor and then back before going to the right node.
    neighbors = {n for n in graph.disorient().neighbors(middle) if n != middle}
    for neighbor in neighbors:
        if (
            _triple_helper(graph, left, middle, neighbor, conditions, sigma)
            and _triple_helper(graph, middle, neighbor, middle, conditions, sigma)
            and _triple_helper(graph, neighbor, middle, right, conditions, sigma)
        ):
            return True
    return False


def _triple_helper(
    graph: NxMixedGraph,
    left: Variable,
    middle: Variable,
    right: Variable,
    conditions: set[Variable],
    sigma: dict[Variable, set[Variable]],
) -> bool:
    return (
        is_collider(graph, left, middle, right, conditions)
        or is_non_collider_left_chain(graph, left, middle, right, conditions, sigma)
        or is_non_collider_right_chain(graph, left, middle, right, conditions, sigma)
        or is_non_collider_fork(graph, left, middle, right, conditions, sigma)
    )


def _has_either_edge(graph: NxMixedGraph, u, v) -> bool:
    return graph.directed.has_edge(u, v) or graph.undirected.has_edge(u, v)


def _only_directed_edge(graph, u, v) -> bool:
    return graph.directed.has_edge(u, v) and not graph.undirected.has_edge(u, v)


def is_collider(
    graph: NxMixedGraph,
    left: Variable,
    middle: Variable,
    right: Variable,
    conditions: set[Variable],
) -> bool:
    """Check if three nodes form a collider under the given conditions.

    :param graph: A mixed graph
    :param left: The first node in the subsequence, denoted as $v_{i-1}$ in the paper
    :param middle: The second node in the subsequence, denoted as $v_i$ in the paper
    :param right: The third node in the subsequence, denoted as $v_{i+1}$ in the paper
    :param conditions: The conditional variables, denoted as $Z$ in the paper
    :return: If the three nodes form a collider
    """
    return (
        _has_either_edge(graph, left, middle)
        and _has_either_edge(graph, right, middle)
        and middle in conditions
    )


def is_non_collider_left_chain(
    graph: NxMixedGraph,
    left: Variable,
    middle: Variable,
    right: Variable,
    conditions: set[Variable],
    sigma: dict[Variable, set[Variable]],
) -> bool:
    r"""Check if three nodes form a non-collider (left chain) given the conditions.

    :param graph: A mixed graph
    :param left: The first node in the subsequence, denoted as $v_{i-1}$ in the paper
    :param middle: The second node in the subsequence, denoted as $v_i$ in the paper
    :param right: The third node in the subsequence, denoted as $v_{i+1}$ in the paper
    :param conditions: The conditional variables, denoted as $Z$ in the paper
    :param sigma: The set of equivalence classes. Can be calculated with
        :func:`get_equivalence_classes`, denoted by $\sigma(v)$ in the paper.
    :return: If the three nodes form a non-collider (left chain) given the conditions.
    """
    return (
        _only_directed_edge(graph, middle, left)
        and _has_either_edge(graph, right, middle)
        and (middle not in conditions or middle in conditions.intersection(sigma[left]))
    )


def is_non_collider_right_chain(
    graph: NxMixedGraph,
    left: Variable,
    middle: Variable,
    right: Variable,
    conditions: set[Variable],
    sigma: dict[Variable, set[Variable]],
) -> bool:
    r"""Check if three nodes form a non-collider (right chain) given the conditions.

    :param graph: A mixed graph
    :param left: The first node in the subsequence, denoted as $v_{i-1}$ in the paper
    :param middle: The second node in the subsequence, denoted as $v_i$ in the paper
    :param right: The third node in the subsequence, denoted as $v_{i+1}$ in the paper
    :param conditions: The conditional variables, denoted as $Z$ in the paper
    :param sigma: The set of equivalence classes. Can be calculated with
        :func:`get_equivalence_classes`, denoted by $\sigma(v)$ in the paper.
    :return: If the three nodes form a non-collider (right chain) given the conditions.
    """
    return (
        _has_either_edge(graph, left, middle)
        and _only_directed_edge(graph, middle, right)
        and (middle not in conditions or middle in conditions.intersection(sigma[right]))
    )


def is_non_collider_fork(
    graph: NxMixedGraph,
    left: Variable,
    middle: Variable,
    right: Variable,
    conditions: set[Variable],
    sigma: dict[Variable, set[Variable]],
) -> bool:
    r"""Check if three nodes form a non-collider (fork) given the conditions.

    :param graph: A mixed graph
    :param left: The first node in the subsequence, denoted as $v_{i-1}$ in the paper
    :param middle: The second node in the subsequence, denoted as $v_i$ in the paper
    :param right: The third node in the subsequence, denoted as $v_{i+1}$ in the paper
    :param conditions: The conditional variables, denoted as $Z$ in the paper
    :param sigma: The set of equivalence classes. Can be calculated with
        :func:`get_equivalence_classes`, denoted by $\sigma(v)$ in the paper.
    :return: If the three nodes form a non-collider (fork) given the conditions.
    """
    a = _only_directed_edge(graph, middle, left)
    b = _only_directed_edge(graph, middle, right)
    c = middle not in conditions
    d = middle in conditions.intersection(sigma[left]).intersection(sigma[right])
    return a and b and (c or d)


def get_equivalence_classes(graph: NxMixedGraph) -> dict[Variable, set[Variable]]:
    """Get equivalence classes.

    :param graph: A mixed graph
    :returns: A mapping from variables to their equivalence class,
        defined as the second option from the paper (see below)

    1. The finest/trivial σ-CG structure of
       a mixed graph G is given by σ(v) := {v} for all
       v ∈ V . In this way σ-separation in G coincides with
       the usual notion of d-separation in a d-connection
       graph (d-CG) G (see [19]). We will take this as the
       definition of d-separation and d-CG in the following.
    2. The coarsest σ-CG structure of a mixed graph G is
       given by σ(v) := ScG(v) := AncG(v) ∩ DescG(v)
       w.r.t. the underlying directed graph. Note that the
       definition of strongly connected component totally
       ignores the bi- and undirected edges of the σ-CG.
    """
    return {
        node: graph.ancestors_inclusive(node).intersection(graph.descendants_inclusive(node))
        for node in graph.nodes()
    }



# -*- coding: utf-8 -*-

"""Graph data structures."""

from __future__ import annotations

import itertools as itt
import json
import warnings
from dataclasses import dataclass, field
from typing import (
    Any,
    Collection,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

import networkx as nx
from networkx.classes.reportviews import NodeView
from networkx.utils import open_file

from .dsl import CounterfactualVariable, Intervention, Variable, vmap_adj, vmap_pairs

__all__ = [
    "NxMixedGraph",
    "CausalEffectGraph",
    "DEFULT_PREFIX",
    "DEFAULT_TAG",
    "set_latent",
]

CausalEffectGraph = Any

#: The default key in a latent variable DAG represented as a :class:`networkx.DiGraph`
#: for nodes that correspond to "latent" variables
DEFAULT_TAG = "hidden"
#: The default prefix for latent variables in a latent variable DAG represented. After the prefix,
#: there will be a number assigned that's incremented during construction.
DEFULT_PREFIX = "u_"


@dataclass
class NxMixedGraph:
    """A mixed graph based on a :class:`networkx.Graph` and a :class:`networkx.DiGraph`.

    Example usage:

    .. code-block:: python

        graph = NxMixedGraph()
        graph.add_directed_edge('X', 'Y')
        graph.add_undirected_edge('X', 'Y')
    """

    #: A directed graph
    directed: nx.DiGraph = field(default_factory=nx.DiGraph)
    #: A undirected graph
    undirected: nx.Graph = field(default_factory=nx.Graph)

    def __eq__(self, other: Any) -> bool:
        """Check for equality of nodes, directed edges, and undirected edges."""
        return (
            isinstance(other, NxMixedGraph)
            and self.nodes() == other.nodes()
            and (self.directed.edges() == other.directed.edges())
            and (self.undirected.edges() == other.undirected.edges())
        )

    def __iter__(self) -> Iterable[Variable]:
        """Iterate over nodes in the graph."""
        return iter(self.directed)

    def __len__(self) -> int:
        """Count the nodes in the graph."""
        return len(self.directed)

    def __contains__(self, item: Variable) -> bool:
        """Check if the given item is a node in the graph."""
        return item in self.directed

    def is_counterfactual(self) -> bool:
        """Check if this is a counterfactual graph."""
        return any(isinstance(n, CounterfactualVariable) for n in self.nodes())

    def raise_on_counterfactual(self) -> None:
        """Raise an error if this is a counterfactual graph.

        :raises ValueError: if this graph is a counterfactual graph
        """
        if self.is_counterfactual():
            raise ValueError("This operation is not available for counterfactual graphs")

    def add_node(self, n: Variable) -> None:
        """Add a node."""
        n = Variable.norm(n)
        self.directed.add_node(n)
        self.undirected.add_node(n)

    def add_directed_edge(self, u: Union[str, Variable], v: Union[str, Variable], **attr) -> None:
        """Add a directed edge from u to v."""
        u = Variable.norm(u)
        v = Variable.norm(v)
        self.directed.add_edge(u, v, **attr)
        self.undirected.add_node(u)
        self.undirected.add_node(v)

    def add_undirected_edge(self, u: Variable, v: Variable, **attr) -> None:
        """Add an undirected edge between u and v."""
        u = Variable.norm(u)
        v = Variable.norm(v)
        self.undirected.add_edge(u, v, **attr)
        self.directed.add_node(u)
        self.directed.add_node(v)

    def nodes(self) -> NodeView[Variable]:
        """Get the nodes in the graph."""
        return self.directed.nodes()

    def to_admg(self):
        """Get an ananke ADMG."""
        self.raise_on_counterfactual()
        from ananke.graphs import ADMG

        # update the way stringification happens so this
        # can support arbitrary variables, like counterfactuals
        return ADMG(
            vertices=[n.name for n in self.nodes()],
            di_edges=[(u.name, v.name) for u, v in self.directed.edges()],
            bi_edges=[(u.name, v.name) for u, v in self.undirected.edges()],
        )

    @classmethod
    def from_admg(cls, admg) -> NxMixedGraph:
        """Create from an ananke ADMG."""
        return cls.from_str_edges(
            nodes=admg.vertices,
            directed=admg.di_edges,
            undirected=admg.bi_edges,
        )

    def to_latent_variable_dag(
        self,
        *,
        prefix: Optional[str] = None,
        start: int = 0,
        tag: Optional[str] = None,
    ) -> nx.DiGraph:
        """Create a labeled DAG where bi-directed edges are assigned as nodes upstream of their two incident nodes.

        :param prefix: The prefix for latent variables. If none, defaults to :data:`y0.graph.DEFAULT_PREFIX`.
        :param start: The starting number for latent variables (defaults to 0, could be changed to 1 if desired)
        :param tag: The key for node data describing whether it is latent.
            If None, defaults to :data:`y0.graph.DEFAULT_TAG`.
        :return: A latent variable DAG.
        """
        self.raise_on_counterfactual()
        return _latent_dag(
            di_edges=self.directed.edges(),
            bi_edges=self.undirected.edges(),
            prefix=prefix,
            start=start,
            tag=tag,
        )

    @classmethod
    def from_latent_variable_dag(cls, graph: nx.DiGraph, tag: Optional[str] = None) -> NxMixedGraph:
        """Load a labeled DAG."""
        if tag is None:
            tag = DEFAULT_TAG
        if any(tag not in data for data in graph.nodes.values()):
            raise ValueError(f"missing label {tag} in one or more nodes.")

        rv = cls()
        for node, data in graph.nodes.items():
            if data[tag]:
                for a, b in itt.combinations(graph.successors(node), 2):
                    rv.add_undirected_edge(a, b)
            else:
                for child in graph.successors(node):
                    rv.add_directed_edge(node, child)
        return rv

    def to_causaleffect(self) -> CausalEffectGraph:
        """Get a causaleffect R object.

        :returns: A causaleffect R object.

        .. warning:: Appropriate R imports need to be done first for 'causaleffect' and 'igraph'.
        """
        import rpy2.robjects

        return rpy2.robjects.r(self.to_causaleffect_str())

    def joint(self) -> nx.MultiGraph:
        """Return a joint graph."""
        rv = nx.MultiGraph()
        rv.add_nodes_from(self.directed)
        rv.add_edges_from(self.directed.edges)
        rv.add_edges_from(self.undirected.edges)
        return rv

    def draw(
        self, ax=None, title: Optional[str] = None, prog: Optional[str] = None, latex: bool = True
    ) -> None:
        """Render the graph using matplotlib.

        :param ax: Axis to draw on (if none specified, makes a new one)
        :param title: The optional title to show with the graph
        :param prog: The pydot program to use, like dot, neato, osage, etc.
            If none is given, uses osage for small graphs and dot for larger ones.
        :param latex: Parse string variables as y0 if possible to make pretty latex output
        """
        import matplotlib.pyplot as plt

        if prog is None:
            if self.directed.number_of_nodes() > 6:
                prog = "dot"
            else:
                prog = "osage"

        layout = _layout(self, prog=prog)
        u_proxy = nx.DiGraph(self.undirected.edges)
        labels = None if not latex else {node: _get_latex(node) for node in self.directed}

        if ax is None:
            ax = plt.gca()

        # TODO choose sizes based on size of axis
        node_size = 1_500
        node_size_offset = 500
        line_widths = 2
        margins = 0.3
        font_size = 20
        arrow_size = 20
        radius = 0.3

        nx.draw_networkx_nodes(
            self.directed,
            pos=layout,
            node_color="white",
            node_size=node_size,
            edgecolors="black",
            linewidths=line_widths,
            ax=ax,
            margins=margins,
        )
        nx.draw_networkx_labels(
            self.directed, pos=layout, ax=ax, labels=labels, font_size=font_size
        )
        nx.draw_networkx_edges(
            self.directed,
            pos=layout,
            edge_color="black",
            ax=ax,
            node_size=node_size + node_size_offset,
            width=line_widths,
            arrowsize=arrow_size,
        )
        nx.draw_networkx_edges(
            u_proxy,
            pos=layout,
            node_size=node_size + node_size_offset,
            ax=ax,
            style=":",
            width=line_widths,
            connectionstyle=f"arc3, rad={radius}",
            arrowstyle="-",
            edge_color="grey",
        )

        if title:
            ax.set_title(title)
        ax.axis("off")

    @classmethod
    def from_causaleffect(cls, graph) -> NxMixedGraph:
        """Construct an instance from a causaleffect R graph."""
        raise NotImplementedError

    def to_causaleffect_str(self) -> str:
        """Get a string to be imported by R."""
        if not self.directed:
            raise ValueError("graph must have some directed edges")

        formula = ", ".join(f"{u} -+ {v}" for u, v in self.directed.edges())
        if self.undirected:
            formula += "".join(f", {u} -+ {v}, {v} -+ {u}" for u, v in self.undirected.edges())

        rv = f"g <- graph.formula({formula}, simplify = FALSE)"
        for i in range(self.undirected.number_of_edges()):
            idx = 2 * i + self.directed.number_of_edges() + 1
            rv += (
                f'\ng <- set.edge.attribute(graph = g, name = "description",'
                f' index = c({idx}, {idx + 1}), value = "U")'
            )

        return rv

    @classmethod
    def from_edges(
        cls,
        nodes: Optional[Iterable[Variable]] = None,
        directed: Optional[Iterable[Tuple[Variable, Variable]]] = None,
        undirected: Optional[Iterable[Tuple[Variable, Variable]]] = None,
    ) -> NxMixedGraph:
        """Make a mixed graph from a pair of edge lists."""
        if directed is None and undirected is None:
            raise ValueError("must provide at least one of directed/undirected edge lists")
        rv = cls()
        for n in nodes or []:
            rv.add_node(n)
        for u, v in directed or []:
            rv.add_directed_edge(u, v)
        for u, v in undirected or []:
            rv.add_undirected_edge(u, v)
        return rv

    @classmethod
    def from_str_edges(
        cls,
        nodes: Optional[Iterable[str]] = None,
        directed: Optional[Iterable[Tuple[str, str]]] = None,
        undirected: Optional[Iterable[Tuple[str, str]]] = None,
    ) -> NxMixedGraph:
        """Make a mixed graph from a pair of edge lists where nodes are strings."""
        return cls.from_edges(
            nodes=None if nodes is None else [Variable(n) for n in nodes],
            directed=None if directed is None else vmap_pairs(directed),
            undirected=None if undirected is None else vmap_pairs(undirected),
        )

    @classmethod
    def from_adj(
        cls,
        nodes: Optional[Iterable[Variable]] = None,
        directed: Optional[Mapping[Variable, Collection[Variable]]] = None,
        undirected: Optional[Mapping[Variable, Collection[Variable]]] = None,
    ) -> NxMixedGraph:
        """Make a mixed graph from a pair of adjacency lists."""
        rv = cls()
        for n in nodes or []:
            rv.add_node(n)
        for u, vs in (directed or {}).items():
            rv.add_node(u)
            for v in vs:
                rv.add_directed_edge(u, v)
        for u, vs in (undirected or {}).items():
            rv.add_node(u)
            for v in vs:
                rv.add_undirected_edge(u, v)
        return rv

    @classmethod
    def from_str_adj(
        cls,
        nodes: Optional[Iterable[str]] = None,
        directed: Optional[Mapping[str, Collection[str]]] = None,
        undirected: Optional[Mapping[str, Collection[str]]] = None,
    ) -> NxMixedGraph:
        """Make a mixed graph from a pair of adjacency lists of strings."""
        return cls.from_adj(
            nodes=None if nodes is None else [Variable(n) for n in nodes],
            directed=None if directed is None else vmap_adj(directed),
            undirected=None if undirected is None else vmap_adj(undirected),
        )

    @classmethod
    @open_file(1)
    def from_causalfusion_path(cls, file) -> NxMixedGraph:
        """Load a graph from a CausalFusion JSON file."""
        return cls.from_causalfusion_json(json.load(file))

    @classmethod
    def from_causalfusion_json(cls, data: Mapping[str, Any]) -> NxMixedGraph:
        """Load a graph from a CausalFusion JSON object."""
        rv = cls()
        for edge in data["edges"]:
            u, v = edge["from"], edge["to"]
            if edge["type"] == "directed":
                rv.add_directed_edge(u, v)
            elif edge["type"] == "bidirected":
                rv.add_undirected_edge(u, v)
            else:
                raise ValueError(f'unhandled edge type: {edge["type"]}')
        return rv

    def subgraph(self, vertices: Union[Variable, Iterable[Variable]]) -> NxMixedGraph:
        """Return a subgraph given a set of vertices.

        :param vertices: a subset of nodes
        :returns: A NxMixedGraph subgraph
        """
        vertices = _ensure_set(vertices)
        return self.from_edges(
            nodes=vertices,
            directed=_include_adjacent(self.directed, vertices),
            undirected=_include_adjacent(self.undirected, vertices),
        )

    def remove_in_edges(self, vertices: Union[Variable, Iterable[Variable]]) -> NxMixedGraph:
        """Return a mutilated graph given a set of interventions.

        :param vertices: a subset of nodes from which to remove incoming edges
        :returns: A NxMixedGraph subgraph
        """
        vertices = _ensure_set(vertices)
        return self.from_edges(
            nodes=vertices,
            directed=_exclude_target(self.directed, vertices),
            undirected=_exclude_adjacent(self.undirected, vertices),
        )

    def get_intervened_ancestors(self, interventions, outcomes) -> Set[Variable]:
        """Get the ancestors of outcomes in a graph that has been intervened on.

        :param interventions: a set of interventions in the graph
        :param outcomes: a set of outcomes in the graph
        :returns: Set of nodes
        """
        return self.remove_in_edges(interventions).ancestors_inclusive(outcomes)

    def get_no_effect_on_outcomes(self, interventions, outcomes) -> Set[Variable]:
        """Find nodes in the graph which have no effect on the outcomes.

        :param interventions: a set of interventions in the graph
        :param outcomes: a set of outcomes in the graph
        :returns: Set of nodes
        """
        return self.nodes() - interventions - self.get_intervened_ancestors(interventions, outcomes)

    def remove_nodes_from(self, vertices: Union[Variable, Iterable[Variable]]) -> NxMixedGraph:
        """Return a subgraph that does not contain any of the specified vertices.

        :param vertices: a set of nodes to remove from graph
        :returns: A NxMixedGraph subgraph
        """
        vertices = _ensure_set(vertices)
        return self.from_edges(
            nodes=self.nodes() - vertices,
            directed=_exclude_adjacent(self.directed, vertices),
            undirected=_exclude_adjacent(self.undirected, vertices),
        )

    def remove_out_edges(self, vertices: Union[Variable, Iterable[Variable]]) -> NxMixedGraph:
        """Return a subgraph that does not have any outgoing edges from any of the given vertices.

        :param vertices: a set of nodes whose outgoing edges get removed from the graph
        :returns: NxMixedGraph subgraph
        """
        vertices = _ensure_set(vertices)
        return self.from_edges(
            nodes=self.nodes(),
            directed=_exclude_source(self.directed, vertices),
            undirected=self.undirected.edges(),
        )

    def ancestors_inclusive(self, sources: Union[Variable, Iterable[Variable]]) -> set[Variable]:
        """Ancestors of a set include the set itself."""
        sources = _ensure_set(sources)
        return _ancestors_inclusive(self.directed, sources)

    def descendants_inclusive(self, sources: Union[Variable, Iterable[Variable]]) -> set[Variable]:
        """Descendants of a set include the set itself."""
        sources = _ensure_set(sources)
        return _descendants_inclusive(self.directed, sources)

    def topological_sort(self) -> Iterable[Variable]:
        """Get a topological sort from the directed component of the mixed graph."""
        return nx.topological_sort(self.directed)

    def get_c_components(self) -> list[frozenset[Variable]]:
        """Get the co-components (i.e., districts) in the undirected portion of the graph."""
        warnings.warn("use NxMixedGraph.districts()", DeprecationWarning, stacklevel=2)
        return list(self.districts())

    def districts(self) -> set[frozenset[Variable]]:
        """Get the districts."""
        return {frozenset(c) for c in nx.connected_components(self.undirected)}

    def get_district(self, node: Variable) -> frozenset[Variable]:
        """Get the district the node is in."""
        for district in self.districts():
            if node in district:
                return district
        raise KeyError(f"{node} not found in graph")

    def is_connected(self) -> bool:
        """Return if there is only a single connected component in the undirected graph."""
        return nx.is_connected(self.undirected)

    def intervene(self, variables: Set[Intervention]) -> NxMixedGraph:
        """Intervene on the given variables.

        :param variables: A set of interventions
        :returns: A graph that has been intervened on the given variables, with edges into the intervened nodes removed
        """
        return self.from_edges(
            nodes=[node.intervene(variables) for node in self.nodes()],
            directed=[
                (u.intervene(variables), v.intervene(variables))
                for u, v in self.directed.edges()
                if _node_not_an_intervention(v, variables)
            ],
            undirected=[
                (u.intervene(variables), v.intervene(variables))
                for u, v in self.undirected.edges()
                if _node_not_an_intervention(u, variables)
                and _node_not_an_intervention(v, variables)
            ],
        )

    def get_markov_pillow(self, nodes: Collection[Variable]) -> Set[Variable]:
        """For each district, intervene on the domain of each parent not in the district."""
        parents_of_district: Set[Variable] = set()
        for node in nodes:
            parents_of_district |= set(self.directed.predecessors(node))
        return parents_of_district - set(nodes)

    def get_markov_blanket(self, nodes: Union[Variable, Iterable[Variable]]) -> Set[Variable]:
        """Get the Markov blanket for a set of nodes.

        The Markov blanket in a directed graph is the union of the parents, children,
        and parents of children of a given node.

        :param nodes: A node or nodes to get the Markov blanket from
        :return: A set of variables comprising the Markov blanket
        """
        if isinstance(nodes, Variable):
            nodes = {nodes}
        else:
            nodes = set(nodes)
        blanket = set()
        for node in nodes:
            blanket.update(self.directed.predecessors(node))
            for successor in self.directed.successors(node):
                blanket.add(successor)
                blanket.update(self.directed.predecessors(successor))
        return blanket.difference(nodes)

    def disorient(self) -> nx.Graph:
        """Return a graph with all edges converted to a flat undirected graph."""
        rv = nx.Graph()
        rv.add_nodes_from(self.nodes())
        rv.add_edges_from(self.directed.edges())
        rv.add_edges_from(self.undirected.edges())
        return rv

    def pre(
        self,
        nodes: Union[Variable, Iterable[Variable]],
        topological_sort_order: Optional[Sequence[Variable]] = None,
    ) -> list[Variable]:
        """Find all nodes prior to the given set of nodes under a topological sort order.

        :param nodes: iterable of nodes.
        :param topological_sort_order: A valid topological sort order. If none given, calculates from the graph.
        :return: list corresponding to the order up until the given nodes.
            This does not include any of the nodes from the query.
        """
        if not topological_sort_order:
            topological_sort_order = list(self.topological_sort())
        node_set = _ensure_set(nodes)
        pre = []
        for node in topological_sort_order:
            if node in node_set:
                break
            pre.append(node)
        return pre


def _node_not_an_intervention(node: Variable, interventions: Set[Intervention]) -> bool:
    """Confirm that node is not an intervention."""
    if isinstance(node, (Intervention, CounterfactualVariable)):
        raise TypeError(
            "this shouldn't happen since the graph should not have interventions as nodes"
        )
    return (+node not in interventions) and (-node not in interventions)


def _ancestors_inclusive(graph: nx.DiGraph, sources: set[Variable]) -> set[Variable]:
    ancestors = set(
        itt.chain.from_iterable(nx.algorithms.dag.ancestors(graph, source) for source in sources)
    )
    return sources | ancestors


def _descendants_inclusive(graph: nx.DiGraph, sources: set[Variable]) -> set[Variable]:
    descendants = set(
        itt.chain.from_iterable(nx.algorithms.dag.descendants(graph, source) for source in sources)
    )
    return sources | descendants


def _include_adjacent(
    graph: nx.Graph, vertices: set[Variable]
) -> Collection[Tuple[Variable, Variable]]:
    vertices = _ensure_set(vertices)
    return [(u, v) for u, v in graph.edges() if u in vertices and v in vertices]


def _exclude_source(
    graph: nx.Graph, vertices: set[Variable]
) -> Collection[Tuple[Variable, Variable]]:
    return [(u, v) for u, v in graph.edges() if u not in vertices]


def _exclude_target(
    graph: nx.Graph, vertices: set[Variable]
) -> Collection[Tuple[Variable, Variable]]:
    return [(u, v) for u, v in graph.edges() if v not in vertices]


def _exclude_adjacent(
    graph: nx.Graph, vertices: set[Variable]
) -> Collection[Tuple[Variable, Variable]]:
    return [(u, v) for u, v in graph.edges() if u not in vertices and v not in vertices]


def _latent_dag(
    di_edges: Iterable[Tuple[Variable, Variable]],
    bi_edges: Iterable[Tuple[Variable, Variable]],
    *,
    prefix: Optional[str] = None,
    start: int = 0,
    tag: Optional[str] = None,
) -> nx.DiGraph:
    """Create a labeled DAG where bi-directed edges are assigned as nodes upstream of their two incident nodes.

    :param di_edges: A list of directional edges
    :param bi_edges: A list of bi-directional edges
    :param prefix: The prefix for latent variables. If none, defaults to :data:`y0.graph.DEFAULT_PREFIX`.
    :param start: The starting number for latent variables (defaults to 0, could be changed to 1 if desired)
    :param tag: The key for node data describing whether it is latent.
        If None, defaults to :data:`y0.graph.DEFAULT_TAG`.
    :return: A latent variable DAG.
    """
    if tag is None:
        tag = DEFAULT_TAG
    if prefix is None:
        prefix = DEFULT_PREFIX

    str_di_edges = [(u.name, v.name) for u, v in di_edges]
    str_bi_edges = [(u.name, v.name) for u, v in bi_edges]

    rv = nx.DiGraph()
    rv.add_nodes_from(itt.chain.from_iterable(str_bi_edges))
    rv.add_edges_from(str_di_edges)
    nx.set_node_attributes(rv, False, tag)
    for i, (u, v) in enumerate(sorted(str_bi_edges), start=start):
        latent_node = f"{prefix}{i}"
        rv.add_node(latent_node, **{tag: True})
        rv.add_edge(latent_node, u)
        rv.add_edge(latent_node, v)
    return rv


def set_latent(
    graph: nx.DiGraph,
    latent_nodes: Union[Variable, Iterable[Variable]],
    tag: Optional[str] = None,
) -> None:
    """Quickly set the latent variables in a graph."""
    if tag is None:
        tag = DEFAULT_TAG
    if isinstance(latent_nodes, Variable):
        latent_nodes = [latent_nodes]

    latent_nodes = set(latent_nodes)
    for node, data in graph.nodes(data=True):
        data[tag] = node in latent_nodes


def _get_latex(node) -> str:
    if isinstance(node, str):
        from y0.parser import parse_y0

        try:
            expr = parse_y0(node)
        except Exception:
            return node
        else:
            return expr._repr_latex_()

    from y0.dsl import Variable

    if isinstance(node, Variable):
        return node._repr_latex_()
    raise TypeError


def _ensure_set(vertices: Union[Variable, Iterable[Variable]]) -> set[Variable]:
    rv = {vertices} if isinstance(vertices, Variable) else set(vertices)
    if any(isinstance(v, Intervention) for v in rv):
        raise TypeError("can not use interventions here")
    return rv


def _layout(self, prog):
    joint = self.joint()
    try:
        layout = nx.nx_agraph.pygraphviz_layout(joint, prog=prog)
    except ImportError:
        pass
    else:
        return layout
    try:
        layout = nx.nx_pydot.pydot_layout(joint, prog=prog)
    except ImportError:
        pass
    else:
        return layout
    return nx.spring_layout(joint)


def is_a_fixable(graph: NxMixedGraph, treatments: Union[Variable, Collection[Variable]]) -> bool:
    """Check if the treatments are a-fixable.

    A treatment is said to be a-fixable if it can be fixed by removing a single directed edge from the graph.
    In other words, a treatment is a-fixable if it has exactly one descendant in its district.

    This code was adapted from :mod:`ananke` ananke code at:
    https://gitlab.com/causal/ananke/-/blob/dev/ananke/estimation/counterfactual_mean.py?ref_type=heads#L58-65

    :param graph: A NxMixedGraph
    :param treatments: A list of treatments
    :raises NotImplementedError: a-fixability on multiple treatments is an open research question
    :returns: bool
    """
    if not isinstance(treatments, Variable):
        raise NotImplementedError(
            "a-fixability on multiple treatments is an open research question"
        )
    descendants = graph.descendants_inclusive(treatments)
    descendants_in_district = graph.get_district(treatments).intersection(descendants)
    return 1 == len(descendants_in_district)


def is_p_fixable(graph: NxMixedGraph, treatments: Union[Variable, Collection[Variable]]) -> bool:
    """Check if the treatments are p-fixable.

    This code was adapted from :mod:`ananke` ananke code at:
    https://gitlab.com/causal/ananke/-/blob/dev/ananke/estimation/counterfactual_mean.py?ref_type=heads#L85-92

    :param graph: A NxMixedGraph
    :param treatments: A list of treatments
    :raises NotImplementedError: p-fixability on multiple treatments is an open research question
    :returns: bool
    """
    if not isinstance(treatments, Variable):
        raise NotImplementedError(
            "p-fixability on multiple treatments is an open research question"
        )
    children = set(graph.directed.successors(treatments))
    children_in_district = graph.get_district(treatments).intersection(children)
    return 0 == len(children_in_district)


def is_markov_blanket_shielded(graph: NxMixedGraph) -> bool:
    """Check if the ADMG is a Markov blanket shielded.

    Being Markov blanket (Mb) shielded means that two vertices are non-adjacent
    only when they are absent from each others' Markov blankets.

    This code was adapted from :mod:`ananke` ananke code at:
    https://gitlab.com/causal/ananke/-/blob/dev/ananke/graphs/admg.py?ref_type=heads#L381-403

    :param graph: A NxMixedGraph
    :returns: bool
    """
    # Iterate over all pairs of vertices
    for u, v in itt.combinations(graph.nodes(), 2):
        # Check if the pair is not adjacent
        if not (
            any(
                [
                    graph.directed.has_edge(u, v),
                    graph.directed.has_edge(v, u),
                    graph.undirected.has_edge(u, v),
                ]
            )
        ):
            # If one is in the Markov blanket of the other, then it is not mb-shielded
            if _markov_blanket_overlap(graph, u, v):
                return False
    return True


def get_district_and_predecessors(
    graph: NxMixedGraph,
    nodes: Iterable[Variable],
    topological_sort_order: Optional[Sequence[Variable]] = None,
):
    """Get the union of district, predecessors and predecessors of district for a given set of nodes.

    This code was adapted from :mod:`ananke` ananke code at:
    https://gitlab.com/causal/ananke/-/blob/dev/ananke/graphs/admg.py?ref_type=heads#L96-117

    :param graph: A NxMixedGraph
    :param nodes: List of nodes
    :param topological_sort_order: A valid topological sort order

    :return: Set corresponding to union of district, predecessors and predecessors of district of a given set of nodes
    """
    if not topological_sort_order:
        topological_sort_order = list(graph.topological_sort())

    # Get the subgraph corresponding to the nodes and nodes prior to them
    pre = graph.pre(nodes, topological_sort_order)
    sub_graph = graph.subgraph(pre + list(nodes))

    result: Set[Variable] = set()
    for node in nodes:
        result.update(sub_graph.get_district(node))
    for node in result.copy():
        result.update(sub_graph.directed.predecessors(node))
    return result - set(nodes)


def _markov_blanket_overlap(graph: NxMixedGraph, u: Variable, v: Variable) -> bool:
    return u in get_district_and_predecessors(graph, [v]) or v in get_district_and_predecessors(
        graph, [u]
    )



"""Utilities for mutation functionality."""

from y0.dsl import Expression, Fraction, Probability, Product, QFactor, Sum

__all__ = [
    "Applier",
]


class Applier:
    """A class for building recursive mutation strategies."""

    def apply_expression(self, expression: Expression) -> Expression:
        """Mutate a generic expression."""
        if isinstance(expression, Sum):
            return self.apply_sum(expression)
        if isinstance(expression, Product):
            return self.apply_product(expression)
        if isinstance(expression, Fraction):
            return self.apply_fraction(expression)
        if isinstance(expression, Probability):
            return self.apply_probability(expression)
        if isinstance(expression, QFactor):
            return self.apply_q(expression)
        return expression

    def apply_sum(self, expression: Sum) -> Expression:
        """Mutate a sum."""
        return Sum(
            expression=self.apply_expression(expression.expression), ranges=expression.ranges
        )

    def apply_product(self, expression: Product) -> Expression:
        """Mutate a product."""
        return Product.safe(self.apply_expression(e) for e in expression.expressions)

    def apply_fraction(self, expression: Fraction) -> Expression:
        """Mutate a fraction."""
        return Fraction(
            numerator=self.apply_expression(expression.numerator),
            denominator=self.apply_expression(expression.denominator),
        )

    def apply_q(self, expression: QFactor) -> Expression:
        """Mutate a probability."""
        return expression

    def apply_probability(self, expression: Probability) -> Expression:
        """Mutate a probability."""
        return expression



"""Contraction functions."""

from operator import attrgetter

from y0.dsl import Distribution, Expression, Fraction, Probability
from y0.mutate.utils import Applier

__all__ = ["contract", "recursive_contract"]


def contract(expression: Expression) -> Expression:
    """Contract a fraction.

    :param expression: A fraction
    :returns: A bayesian compressed probability, if possible

    >>> from y0.dsl import P, A, B, C, D
    >>> assert P(A | B) == contract(P(A, B) / P(B))
    >>> assert P(A | B, C) == contract(P(A, B, C) / P(B, C))
    >>> assert P(B, C | A) == contract(P(A, B, C) / P(A))
    >>> assert P(A, B, C) / P(D) == contract(P(A, B, C) / P(D))
    """
    if not (
        isinstance(expression, Fraction)
        and isinstance(expression.numerator, Probability)
        and isinstance(expression.denominator, Probability)
        and not expression.numerator.parents
        and not expression.denominator.parents
        and set(expression.denominator.children).issubset(expression.numerator.children)
    ):
        return expression
    children = set(expression.numerator.children).difference(expression.denominator.children)
    parents = set(expression.numerator.children).intersection(expression.denominator.children)
    return expression.numerator._new(
        Distribution(
            children=tuple(sorted(children, key=attrgetter("name"))),
            parents=tuple(sorted(parents, key=attrgetter("name"))),
        )
    )


class _Contracter(Applier):
    def apply_fraction(self, expression: Fraction) -> Expression:
        """Contract a fraction."""
        return contract(expression)


def recursive_contract(e: Expression) -> Expression:
    """Recursively contract fractions."""
    return _Contracter().apply_expression(e)



# -*- coding: utf-8 -*-

"""Utilities for parallel world graphs and counterfactual graphs."""

from itertools import combinations
from typing import FrozenSet, Iterable, Optional, Set, Tuple, cast

from y0.dsl import CounterfactualVariable, Event, Intervention, Variable
from y0.graph import NxMixedGraph

__all__ = [
    "has_same_function",
    "extract_interventions",
    "is_pw_equivalent",
    "merge_pw",
    "make_counterfactual_graph",
    "make_parallel_worlds_graph",
    "is_not_self_intervened",
]


class World(FrozenSet[Intervention]):
    """A set of interventions corresponding to a "world"."""

    def __contains__(self, item) -> bool:
        if not isinstance(item, Intervention):
            raise TypeError(
                f"can not check if non-intervention is in a world: ({type(item)}) {item}"
            )
        return super().__contains__(item)


Worlds = Set[World]


def has_same_confounders(graph: NxMixedGraph, a: Variable, b: Variable) -> bool:
    """Check if all confounders of the two nodes are the same."""
    no_undirected_edges = (
        0 == len(list(graph.undirected.edges(a))) == len(list(graph.undirected.edges(b)))
    )
    return graph.undirected.has_edge(a, b) or no_undirected_edges


def has_same_function(node1: Variable, node2: Variable) -> bool:
    """Check if the two nodes have the same functional mechanism."""
    return node1.get_base() == node2.get_base() and is_not_self_intervened(
        node1
    ) == is_not_self_intervened(node2)


def nodes_attain_same_value(graph: NxMixedGraph, event: Event, a: Variable, b: Variable) -> bool:
    """Check if the two nodes attain the same value."""
    if a == b:
        return True
    if not has_same_confounders(graph, a, b):
        return False
    elif a.get_base() != b.get_base():
        return False
    elif a in event and b in event:
        if event[a] != event[b]:
            return False  # D and D @ -d  events = {D: -d}
        else:
            return True
    elif a in event:
        if not isinstance(b, CounterfactualVariable) or event[a] not in b.interventions:
            return False
        else:
            return True
    elif b in event:
        if not isinstance(a, CounterfactualVariable) or event[b] not in a.interventions:
            return False
        else:
            return True
    elif isinstance(a, CounterfactualVariable) or isinstance(b, CounterfactualVariable):
        return False
    return True


def parents_attain_same_values(graph: NxMixedGraph, event: Event, a: Variable, b: Variable) -> bool:
    """Check if the parents of the nodes attain the same value."""
    if not has_same_confounders(graph, a, b):
        return False
    parents_a, parents_b = set(graph.directed.predecessors(a)), set(graph.directed.predecessors(b))
    if parents_a == parents_b:
        return True
    remainder_a, remainder_b = parents_a - parents_b, parents_b - parents_a
    if len(remainder_a) != len(remainder_b):
        return False
    return all(
        nodes_attain_same_value(graph, event, parent_a, parent_b)
        for parent_a, parent_b in zip(
            sorted(remainder_a, key=lambda x: x.get_base()),
            sorted(remainder_b, key=lambda x: x.get_base()),
        )
    )


def is_not_self_intervened(node: Variable) -> bool:
    """Check if the node is not self-intervened."""
    return not isinstance(node, CounterfactualVariable) or (
        +(node.get_base()) not in node.interventions
        and -(node.get_base()) not in node.interventions
    )


def extract_interventions(variables: Iterable[Variable]) -> Worlds:
    """Extract the set of interventions for each counterfactual variable that corresponds to a world."""
    return set(
        World(variable.interventions)
        for variable in variables
        if isinstance(variable, CounterfactualVariable)
    )


def is_pw_equivalent(graph: NxMixedGraph, event: Event, node1: Variable, node2: Variable) -> bool:
    r"""Check if two nodes in a parallel worlds graph are equivalent.

    :param graph: A parallel worlds graph
    :param event: A dictionary of variables and variables values
    :param node1: A node in the graph
    :param node2: Another node in the graph
    :returns: If the two nodes are equivalent under the parallel worlds assumption

    Let :math:`M` be a model inducing :math:`G` containing variables
    :math:`\alpha`, :math:`\beta` with the following properties:

    * :math:`\alpha` and :math:`\beta` have the same domain of values.
    * There is a bijection :math:`f` from :math:`Pa(\alpha)` to :math:`Pa(\beta)`
      such that a parent  :math:`\gamma` and :math:`f(\gamma)` have the same domain
      of values.
    * The functional mechanisms of :math:`\alpha` and :math:`\beta` are the same
      (except whenever the function for :math:`\alpha` uses the parent :math:`\gamma`,
      the corresponding function for :math:`\beta` uses :math:`f(\gamma)`).

    Assume an observable variable set :math:`\mathbf{Z}` was observed to attain values
    :math:`z` in :math:`M_\mathbf{x}` , the submodel obtained from :math:`M` by forcing
    another observable variable set :math:`\mathbf{X}` to attain values :math:`\mathbf{x}`.
    Assume further that for each  :math:`\gamma \in Pa(\alpha)`, either
    :math:`f(\gamma) =  \gamma`, or  :math:`\gamma` and :math:`f(\gamma)` attain the
    same values (whether by observation or intervention). Then :math:`\alpha` and
    :math:`\beta` are the same random variable in :math:`M_\mathbf{x}` with observations
    :math:`\mathbf{z}`.

    Lemma 24 from Ilya Shpitser and Judea Pearl. 2008.
    Complete Identification Methods for the Causal Hierarchy.
    Journal of Machine Learning Research (2008).
    """
    # Rather than all n choose 2 combinations, we can restrict ourselves to the original
    # graph variables and their counterfactual versions
    assert (node1 in graph.nodes()) and (node2 in graph.nodes()), "Nodes must be in the graph"
    return (
        has_same_function(node1, node2)
        and parents_attain_same_values(graph, event, node1, node2)
        and nodes_have_same_domain_of_values(graph, event, node1, node2)
    )


def nodes_have_same_domain_of_values(
    graph: NxMixedGraph, event: Event, a: Variable, b: Variable
) -> bool:
    """Check if the nodes have the same domain of values."""
    if not has_same_confounders(graph, a, b):
        return False
    if a.get_base() != b.get_base():
        return False
    if is_not_self_intervened(a) and is_not_self_intervened(b):
        return True
    if is_not_self_intervened(a) or is_not_self_intervened(b):
        return False
    if value_of_self_intervention(a) == value_of_self_intervention(b):
        return True
    return False


def value_of_self_intervention(a: Variable) -> Optional[Intervention]:
    """Get the value of the self-intervention."""
    if not isinstance(a, CounterfactualVariable):
        return None
    base = a.get_base()
    if +base in a.interventions:
        return cast(Intervention, +base)
    elif -base in a.interventions:
        return cast(Intervention, -base)
    return None


def merge_pw(
    graph: NxMixedGraph, node1: Variable, node2: Variable
) -> Tuple[NxMixedGraph, Variable, Variable]:
    r"""Merge node1 and node2 and return the reduced graph and query.

    :param graph: A parallel worlds graph
    :param node1: A node in the graph
    :param node2: Another node in the graph
    :returns: A reduced graph

    Let :math:`M_\mathbf{x}` be a submodel derived from :math:`M` with set :math:`\mathbf{Z}`
    observed to attain values :math:`\mathbf{z}`, such that Lemma 24 holds for :math:`\alpha`;
    :math:`\beta`. Let :math:`M'` be a causal model obtained from :math:`M` by merging
    :math:`\alpha`; :math:`\beta` into a new node :math:`\omega`, which inherits all parents
    and the functional mechanism of :math:`\alpha`. All children of
    :math:`\alpha`; :math:`\beta` in :math:`M'` become children of :math:`\omega`. Then
    :math:`M_\mathbf{x},  M'_\mathbf{x} agree on any distribution consistent with :math:`z`
    being observed.

    Lemma 25 from Ilya Shpitser and Judea Pearl. 2008.
    Complete Identification Methods for the Causal Hierarchy.
    Journal of Machine Learning Research (2008).
    """
    # If a we are going to merge two nodes, we want to keep the factual variable.
    if isinstance(node1, CounterfactualVariable) and not isinstance(node2, CounterfactualVariable):
        node1, node2 = node2, node1
    elif not isinstance(node1, CounterfactualVariable) and isinstance(
        node2, CounterfactualVariable
    ):
        pass
    else:  # both are counterfactual or both are factual, so keep the variable with the lower name
        node1, node2 = sorted([node1, node2])
    directed = [(u, v) for u, v in graph.directed.edges() if node2 not in (u, v)]
    directed += [(node1, v) for u, v in graph.directed.edges() if node2 == u]
    # directed += [(u, node1) for u, v in graph.directed.edges() if node2 == v]
    undirected = [frozenset({u, v}) for u, v in graph.undirected.edges() if node2 not in (u, v)]
    undirected += [
        frozenset({node1, v}) for u, v in graph.undirected.edges() if node2 == u and node1 != v
    ]
    undirected += [
        frozenset({u, node1}) for u, v in graph.undirected.edges() if node2 == v and node1 != u
    ]
    parents_of_node1 = [u for u, v in graph.directed.edges() if v == node1]
    parents_of_node2_not_node1 = [
        u for u, v in graph.directed.edges() if v == node2 and u not in parents_of_node1
    ]
    return (
        NxMixedGraph.from_edges(
            nodes=[
                node
                for node in graph.nodes()
                if node != node2 and node not in parents_of_node2_not_node1
            ],
            directed=list(set(directed)),
            undirected=[(u, v) for u, v in set(undirected)],
        ),
        node1,
        node2,
    )


def lemma_24_holds(
    cf_graph: NxMixedGraph, event: Event, node: Variable, node_at_interventions: Variable
) -> bool:
    r"""Check if Lemma 24 holds for the given nodes."""
    return (
        (node in cf_graph.nodes())
        and (node_at_interventions in cf_graph.nodes())
        and is_pw_equivalent(cf_graph, event, node, node_at_interventions)
    )


def is_inconsistent(event: Event, node: Variable, node_at_interventions: Variable) -> bool:
    r"""Check if the equivalant nodes (according to lemma 25) have been assigned different values in the event."""
    return (
        (node in event)
        and (node_at_interventions in event)
        and (event[node] != event[node_at_interventions])
    )


def update_event(event: Event, preferred_node: Variable, eliminated_node: Variable) -> Event:
    r"""Update the event to reflect the fact that the preferred node has been merged into the eliminated node."""
    if eliminated_node in event:
        event[preferred_node] = event[eliminated_node]
        del event[eliminated_node]
    return event


def make_counterfactual_graph(
    graph: NxMixedGraph, event: Event
) -> Tuple[NxMixedGraph, Optional[Event]]:
    r"""Make counterfactual graph.

    :param graph: A causal graph :math:`G`
    :param event: A conjunction of counterfactual events :math:`\gamma`
    :returns:
        A counterfactual graph and either a set of new events :math:`\gamma'`
        such that :math:`P(\gamma') = P(\gamma)` or None

    * Construct a submodel graph :math:`G_{\mathbf{x}_i}` for each action
      :math:`do(\mathbf{x}_i)` mentioned in the event :math:`\gamma`. Construct the
      parallel worlds graph :math:`G'` by having all such submodel graphs share their
      corresponding :math:`U` nodes
    * Let :math:`\pi` be a topological ordering of nodes in :math:`G'`, let :math:`\gamma' := \gamma`
    * Apply Lemmas 24 and 25, in order :math:`\pi`, to each observable pair of nodes :math:`\alpha, \beta` in :math:`G`.
      For each :math:`\alpha, \beta` that are the same, do:
        * Let :math:`G'` be modified as specified in Lemma 25
        * Modify :math:`\gamma'` by renaming all occurrences of :math:`\beta` to :math:`\alpha`.
        * If :math:`\mathbf{\alpha} \ne :math:`\mathbf{\beta}`, return :math:`G'`, None
    * Return :math:`(G', \gamma')`, where :math:`An(\gamma')` is the set of nodes in :math:`G'`
      ancestral to nodes corresponding to variables mentioned in :math:`\gamma'`.
    """
    worlds = extract_interventions(event)
    pw_graph = make_parallel_worlds_graph(graph, worlds)
    new_event = dict(event)
    cf_graph = NxMixedGraph.from_edges(
        nodes=pw_graph.nodes(),
        directed=pw_graph.directed.edges(),
        undirected=pw_graph.undirected.edges(),
    )
    for node in graph.topological_sort():
        for world in worlds:
            node_at_interventions = node @ world
            if lemma_24_holds(cf_graph, new_event, node, node_at_interventions):
                cf_graph, preferred_node, eliminated_node = merge_pw(
                    cf_graph, node, node_at_interventions
                )
                if is_inconsistent(new_event, preferred_node, eliminated_node):
                    return cf_graph, None
                new_event = update_event(new_event, preferred_node, eliminated_node)
        if len(worlds) > 1:
            for intervention1, intervention2 in combinations(worlds, 2):
                node_at_intervention1 = node @ intervention1
                node_at_intervention2 = node @ intervention2

                if lemma_24_holds(
                    cf_graph, new_event, node_at_intervention1, node_at_intervention2
                ):
                    cf_graph, preferred_node, eliminated_node = merge_pw(
                        cf_graph, node_at_intervention1, node_at_intervention2
                    )
                    if is_inconsistent(new_event, node_at_intervention1, node_at_intervention2):
                        return cf_graph, None
                    new_event = update_event(new_event, preferred_node, eliminated_node)

    ancestors = cf_graph.ancestors_inclusive(new_event)
    rv_graph = cf_graph.subgraph(ancestors)
    return rv_graph, new_event


def node_not_an_intervention_in_world(world: World, node: Variable) -> bool:
    """Confirm that node is not an intervention in a given world."""
    if isinstance(node, (Intervention, CounterfactualVariable)):
        raise TypeError(
            "this shouldn't happen since the graph should not have interventions as nodes"
        )
    return (+node not in world) and (-node not in world)


def stitch_factual_and_dopplegangers(
    graph: NxMixedGraph, worlds: Worlds
) -> Set[Tuple[Variable, CounterfactualVariable]]:
    """Stitch together a node and its counterfactual doppleganger in each world."""
    return {
        (u, u @ world)
        for world in worlds
        for u in graph.nodes()
        if node_not_an_intervention_in_world(world, u)
    }


def stitch_factual_and_doppleganger_neighbors(
    graph: NxMixedGraph, worlds: Worlds
) -> Set[Tuple[Variable, CounterfactualVariable]]:
    """Stitch together a node with the dopplegangers of its neighbors in each world."""
    return {
        (u, v @ world)
        for world in worlds
        for u in graph.nodes()
        for v in graph.undirected.neighbors(u)
        # Don't add an edge if a variable is intervened on
        if node_not_an_intervention_in_world(world, v)
    }


def stitch_counterfactual_and_dopplegangers(
    graph: NxMixedGraph, worlds: Worlds
) -> Set[Tuple[CounterfactualVariable, CounterfactualVariable]]:
    """Stitch together a counterfactual variable with its doppelganger.

    Unless the counterfactual is intervened upon in one of the worlds.

    :param graph: A NxMixedGraph
    :param worlds: A set of frozensets of interventions
    :returns: A set of undirected edges
    """
    rv = {
        (u @ world_1, u @ world_2)
        for world_1, world_2 in combinations(worlds, 2)
        for u in graph.nodes()
        # Don't add an edge if a variable is intervened on in either world.
        if node_not_an_intervention_in_world(world_1, u)
        and node_not_an_intervention_in_world(world_2, u)
    }
    return _both_ways(rv)


def _both_ways(s):
    rv = set()
    for a, b in s:
        rv.add((b, a))
    return rv


def stitch_counterfactual_and_doppleganger_neighbors(
    graph: NxMixedGraph, worlds: Worlds
) -> Set[Tuple[CounterfactualVariable, CounterfactualVariable]]:
    """Stitch together a counterfactual variable with the dopplegangers of its neighbors in each world."""
    rv = {
        frozenset({u @ world_1, v @ world_2})
        for world_1, world_2 in combinations(worlds, 2)
        for u in graph.nodes()
        for v in graph.undirected.neighbors(u)
        # Don't add an edge if a variable is intervened on in either world.
        if node_not_an_intervention_in_world(world_1, u)
        and node_not_an_intervention_in_world(world_2, v)
    }
    return _both_ways(rv)


def stitch_counterfactual_and_neighbors(
    graph: NxMixedGraph, worlds: Worlds
) -> Set[Tuple[CounterfactualVariable, CounterfactualVariable]]:
    """Stitch together a counterfactual variable with its neighbors in each world."""
    rv = {
        (u @ world, v @ world)
        for world in worlds
        for u in graph.nodes()
        for v in graph.undirected.neighbors(u)
        # Don't add an edge if a variable is intervened on in either world.
        if node_not_an_intervention_in_world(world, u)
        and node_not_an_intervention_in_world(world, v)
    }
    return _both_ways(rv)


def _get_directed_edges(
    graph: NxMixedGraph, worlds: Worlds
) -> Set[Tuple[CounterfactualVariable, CounterfactualVariable]]:
    """Get the directed edges in the parallel worlds graph.

    Except for those where the target node was intervened upon.

    :param graph: A NxMixedGraph
    :param worlds: A set of frozensets of interventions
    :returns: A set of directed edges
    """
    return {
        (u @ world, v @ world)
        for world in worlds
        for u, v in graph.directed.edges()
        if node_not_an_intervention_in_world(world, v)
    }


def make_parallel_worlds_graph(
    graph: NxMixedGraph,
    worlds: Worlds,
) -> NxMixedGraph:
    """Make a parallel worlds graph.

    :param graph: A normal graph
    :param worlds: A set of sets of treatments
    :returns: A combine parallel world graph
    """
    # Get the undirected edges
    undirected: Set[Tuple[Variable, Variable]] = set()
    # get all the undirected edges in all the parallel worlds
    undirected |= stitch_counterfactual_and_neighbors(graph, worlds)
    # Stitch together factual variables with their dopplegangers in other worlds
    undirected |= stitch_factual_and_dopplegangers(graph, worlds)
    # Stitch together factual variables with the dopplegangers of their neighbors in other worlds
    undirected |= stitch_factual_and_doppleganger_neighbors(graph, worlds)

    # Stitch together counterfactual variables with their dopplegangers in other worlds
    if len(worlds) > 1:
        undirected |= stitch_counterfactual_and_dopplegangers(graph, worlds)
        # Stitch together counterfactual variables with the dopplegangers of their neighbors in other worlds
        undirected |= stitch_counterfactual_and_doppleganger_neighbors(graph, worlds)

    nodes = [
        *graph.nodes(),
        *(node @ world for world in worlds for node in graph.nodes()),
    ]
    directed = [*graph.directed.edges(), *_get_directed_edges(graph, worlds)]
    return NxMixedGraph.from_edges(
        nodes=nodes,
        directed=directed,
        undirected=set(graph.undirected.edges()) | undirected,
    )



"""Examples for front-door."""

import numpy as np
import pandas as pd

from y0.dsl import Variable, X, Y, Z

__all__ = [
    "generate_data_for_frontdoor",
]


def _r_exp(x):
    return 1 / (1 + np.exp(x))


def generate_data_for_frontdoor(
    num_samples: int, treatments: dict[Variable, float] | None = None, *, seed: int | None = None
) -> pd.DataFrame:
    """Generate testing data for the front-door case study.

    :param num_samples: The number of samples to generate. Try 1000.
    :param treatments: An optional dictionary of the values to fix each variable to.
    :param seed: An optional random seed for reproducibility purposes
    :returns: A pandas Dataframe with columns corresponding
        to the variable names in the front-door example
    """
    if treatments is None:
        treatments = {}
    generator = np.random.default_rng(seed)

    u = generator.normal(loc=40.0, scale=10.0, size=num_samples)

    beta0_x = -1
    beta_u_to_x = 0.05
    if X in treatments:
        x = np.full(num_samples, treatments[X])
    else:
        p = _r_exp(-beta0_x - u * beta_u_to_x)
        x = generator.binomial(1, p, size=num_samples)

    beta0_z = -1.9
    beta_x_to_z = 0.04
    if Z in treatments:
        z = np.full(num_samples, treatments[Z])
    else:
        z = generator.normal(
            loc=100 * _r_exp(-beta0_z - x * beta_x_to_z), scale=1, size=num_samples
        )

    beta0_y = -1.8
    beta_z_to_y = 0.05
    beta_u_to_y = 0.06
    if Y in treatments:
        y = np.full(num_samples, treatments[Y])
    else:
        y = generator.normal(
            loc=100 * _r_exp(-beta0_y - z * beta_z_to_y - u * beta_u_to_y),
            scale=1,
            size=num_samples,
        )

    return pd.DataFrame({X.name: x, Z.name: z, Y.name: y})



# -*- coding: utf-8 -*-

"""Tests for parallel world graphs and counterfactual graphs."""

from tests.test_algorithm import cases
from y0.algorithm.identify.cg import (
    World,
    _get_directed_edges,
    extract_interventions,
    has_same_confounders,
    has_same_function,
    is_inconsistent,
    is_not_self_intervened,
    is_pw_equivalent,
    make_counterfactual_graph,
    make_parallel_worlds_graph,
    merge_pw,
    node_not_an_intervention_in_world,
    nodes_attain_same_value,
    nodes_have_same_domain_of_values,
    parents_attain_same_values,
    stitch_counterfactual_and_doppleganger_neighbors,
    stitch_counterfactual_and_dopplegangers,
    stitch_counterfactual_and_neighbors,
    stitch_factual_and_doppleganger_neighbors,
    stitch_factual_and_dopplegangers,
    value_of_self_intervention,
)
from y0.dsl import A, B, D, Event, W, X, Y, Z
from y0.examples import (
    figure_9a,
    figure_9b,
    figure_9c,
    figure_9d,
    figure_11a,
    figure_11b,
    figure_11c,
)
from y0.graph import NxMixedGraph

x, y, z, d, w = -X, -Y, -Z, -D, -W


class TestCounterfactualGraph(cases.GraphTestCase):
    """Tests parallel worlds and counterfactual graphs."""

    def test_world(self):
        """Test that a world contains an intervention."""
        with self.assertRaises(TypeError):
            input_world1: World = World([-x])
            3 in input_world1  # noqa

        with self.assertRaises(TypeError):
            input_world1: World = World([3])
            3 in input_world1  # noqa

        input_world2 = World([-x])
        self.assertFalse(+x in input_world2)
        self.assertFalse(+y in input_world2)
        self.assertTrue(-x in input_world2)

    def assert_uedge_set_equal(self, expected, actual):
        """Assert that two undirected edge sets are equal."""
        return self.assertEqual(
            {frozenset(edge) for edge in expected}, {frozenset(edge) for edge in actual}
        )

    def test_make_parallel_worlds(self):
        """Test that parallel worlds graphs are correct."""
        expected = figure_9b.graph
        actual = make_parallel_worlds_graph(figure_9a.graph, {frozenset([-x]), frozenset([-d])})
        expected2 = make_parallel_worlds_graph(figure_9a.graph, {(-x, -z)})
        self.assert_graph_equal(expected, actual)
        self.assert_graph_equal(expected2, make_parallel_worlds_graph(figure_9a.graph, {(-x, -z)}))
        self.assertTrue(Y @ (-z, -x) in expected2.nodes())
        self.assertTrue(Y @ (-x, -z) in expected2.nodes())

    def test_has_same_function(self):
        """Test that two variables have the same value."""
        self.assertTrue(has_same_function(D @ X, D))
        self.assertFalse(has_same_function(D @ D, D))
        self.assertFalse(has_same_function(X, X @ +x))
        self.assertTrue(has_same_function(X @ D, X))
        self.assertFalse(has_same_function(X, D))
        self.assertFalse(has_same_function(X @ ~X, W @ ~X))
        self.assertFalse(has_same_function(X @ ~X, X))
        self.assertTrue(has_same_function(D, D @ ~x))
        self.assertTrue(has_same_function(Z @ ~x, Z))
        self.assertTrue(has_same_function(Z @ ~x, Z @ -d))
        self.assertTrue(has_same_function(Z @ -d, Z))
        self.assertTrue(has_same_function(Z @ (-d, -z), Z @ (-x, -z)))
        self.assertTrue(has_same_function(Z @ (-d, -z), Z @ (-x, +z)))
        self.assertFalse(has_same_function(Z @ (-d, -z), Z @ (-d, +x)))

    def test_nodes_attain_same_value(self):
        """Test that two variables attain the same value."""
        event: Event = {D: -d}
        self.assertTrue(nodes_attain_same_value(figure_11a.graph, event, D, D @ -d))
        self.assertTrue(nodes_attain_same_value(figure_11a.graph, event, D @ -d, D))
        self.assertTrue(
            nodes_attain_same_value(
                NxMixedGraph.from_edges(directed=[(D @ +d, Z @ +d)]), event, D @ +d, D @ +d
            )
        )
        self.assertTrue(nodes_attain_same_value(figure_9b.graph, event, D, D))
        self.assertFalse(nodes_attain_same_value(figure_9b.graph, event, D @ -d, D))
        self.assertFalse(nodes_attain_same_value(figure_9b.graph, event, D, D @ -d))
        self.assertFalse(nodes_attain_same_value(figure_11a.graph, event, Z, Z @ +x))
        self.assertFalse(nodes_attain_same_value(figure_9b.graph, event, D, X))
        self.assertFalse(nodes_attain_same_value(figure_9b.graph, event, X, D))
        self.assertFalse(nodes_attain_same_value(figure_9b.graph, event, D, X @ -d))
        self.assertFalse(nodes_attain_same_value(figure_9b.graph, event, X @ -d, D))
        self.assertFalse(nodes_attain_same_value(figure_9b.graph, event, D, X @ -d))
        self.assertFalse(nodes_attain_same_value(figure_9b.graph, event, D @ -d, D))
        # This should get us onto the return False on line 56
        self.assertFalse(nodes_attain_same_value(figure_9b.graph, event, Y @ -X, X))
        # This one should trigger the False on line 59
        self.assertFalse(nodes_attain_same_value(figure_9b.graph, {Y @ -X: -Y, Y: +Y}, Y @ -X, Y))
        # This one will trigger the True on line 61, but I'm not sure if the event really makes sense

        self.assertTrue(nodes_attain_same_value(figure_9b.graph, {Y @ -X: -Y, Y: -Y}, Y @ -X, Y))
        # I think these two should get the falses on lines 64 and 69
        self.assertFalse(nodes_attain_same_value(figure_9b.graph, {Y @ -X: -Y}, Y @ -X, Y))
        self.assertFalse(nodes_attain_same_value(figure_9b.graph, {Y: +Y}, Y @ -X, Y))
        # These two should get the False on line 73
        self.assertFalse(nodes_attain_same_value(figure_9b.graph, {}, Y @ -X, Y))
        self.assertFalse(nodes_attain_same_value(figure_9b.graph, {}, Y, Y @ -X))
        # These two should get the True on line 74
        self.assertTrue(
            nodes_attain_same_value(NxMixedGraph.from_edges(undirected=[(+Y, -Y)]), {}, +Y, -Y)
        )
        self.assertFalse(nodes_attain_same_value(figure_9b.graph, event, D @ -X, D))

    def test_has_same_confounders(self):
        """Check whether two nodes have the same confounding edges."""
        self.assertFalse(has_same_confounders(figure_9b.graph, D, D))
        self.assertFalse(has_same_confounders(figure_9b.graph, D, D @ -d))
        self.assertFalse(has_same_confounders(figure_9b.graph, D @ -d, D))
        self.assertTrue(has_same_confounders(figure_9b.graph, D @ -x, D))
        self.assertTrue(has_same_confounders(figure_11a.graph, D @ -d, D))
        self.assertTrue(has_same_confounders(figure_9b.graph, D, D @ -x))

    def test_parents_attain_same_values(self):
        """Test that the parents of two nodes attain the same value."""
        graph = figure_9b.graph
        event: Event = {Y @ -x: -y, D: -d, Z @ -d: -z, X: +x}
        self.assertTrue(parents_attain_same_values(figure_11a.graph, event, Z, Z @ -d))
        self.assertTrue(parents_attain_same_values(figure_11a.graph, event, Z, Z @ -x))
        self.assertTrue(parents_attain_same_values(figure_11a.graph, event, Z @ -d, Z @ -x))
        self.assertTrue(parents_attain_same_values(figure_11a.graph, event, D, D @ -d))
        self.assertFalse(parents_attain_same_values(graph, event, Z, Z @ -d))
        self.assertFalse(parents_attain_same_values(figure_9b.graph, event, D, D @ -d))
        self.assertFalse(parents_attain_same_values(figure_9b.graph, event, X, X @ -x))
        self.assertFalse(
            parents_attain_same_values(
                NxMixedGraph.from_edges(directed=[(Z, X), (Z, Y), (W, Y)]), event, X, Y
            )
        )
        self.assertFalse(
            parents_attain_same_values(
                NxMixedGraph.from_edges(
                    directed=[(X @ -x, Y @ -x), (X @ +x, Y @ +x)], undirected=[(Y @ -x, Y @ +x)]
                ),
                {Y @ -x: -y, Y @ +x: +y},
                Y @ -x,
                Y @ +x,
            )
        )

    def test_nodes_have_same_domain_of_values(self):
        """Test that two nodes have the same domain of values."""
        graph = figure_9b.graph
        event = {Y @ -x: -y, X: +x, D: -d, Z @ -d: -z}
        self.assertTrue(nodes_have_same_domain_of_values(graph, event, D, D @ -x))
        self.assertFalse(nodes_have_same_domain_of_values(graph, event, D, D @ -d))
        self.assertFalse(
            nodes_have_same_domain_of_values(
                NxMixedGraph.from_edges(
                    directed=[(X, Y), (X @ +x, Y @ +x)], undirected=[(Y, Y @ +x)]
                ),
                {Y @ +x: +y, Y @ -x: -y},
                X,
                X @ +x,
            )
        )
        self.assertTrue(nodes_have_same_domain_of_values(figure_9b.graph, event, D @ -X, D))
        self.assertFalse(nodes_have_same_domain_of_values(figure_9b.graph, event, X, Y))
        self.assertFalse(
            nodes_have_same_domain_of_values(
                NxMixedGraph.from_edges(
                    directed=[(X @ +x, Y @ +x), (X @ -x, Y @ -x), (X, Y)],
                    undirected=[(Y @ -x, Y @ +x), (Y @ +x, Y), (Y @ -x, Y)],
                ),
                {Y @ -x: -y, Y @ +x: +y},
                X @ -x,
                X @ +x,
            )
        )
        self.assertTrue(
            nodes_have_same_domain_of_values(
                NxMixedGraph.from_edges(
                    directed=[(X @ +x, Y @ +x), (X @ -x, Y @ -x), (X, Y)],
                    undirected=[(Y @ -x, Y @ +x), (Y @ +x, Y), (Y @ -x, Y)],
                ),
                {Y @ -x: -y, Y @ +x: +y},
                X @ +x,
                X @ +x,
            )
        )
        self.assertTrue(
            nodes_have_same_domain_of_values(
                NxMixedGraph.from_edges(
                    directed=[(X @ +x, Y @ +x), (X @ -x, Y @ -x), (X, Y)],
                    undirected=[(Y @ -x, Y @ +x), (Y @ +x, Y), (Y @ -x, Y)],
                ),
                {Y @ -x: +y, Y @ +x: +y},
                Y @ -x,
                Y @ +x,
            )
        )

    def test_value_of_self_intervention(self):
        """Return the value of a self intervention."""
        self.assertIsNone(value_of_self_intervention(X))
        self.assertEqual(+x, value_of_self_intervention(X @ +x))
        self.assertEqual(-x, value_of_self_intervention(X @ (-x, -y, -d)))
        self.assertIsNone(value_of_self_intervention(X @ -y))

    def test_extract_interventions(self):
        """Test that all interventions are extracted."""
        event = {X: +x, D: -d, Z @ -d: -z}
        expected_worlds = {frozenset({-d})}
        self.assertEqual(expected_worlds, extract_interventions(event))

    def test_get_worlds(self):
        """Test that all interventions within each world of a counterfactual conjunction are generated."""
        self.assert_uedge_set_equal(
            {frozenset([-D]), frozenset([~X])}, extract_interventions([Y @ ~X, X, Z @ -D, D])
        )
        self.assert_uedge_set_equal(
            {frozenset([-D]), frozenset([~X, -Z])},
            extract_interventions([Y @ (~X, -Z), X, Z @ -D, D]),
        )

    def test_node_not_an_intervention_in_world(self):
        """Test that a node is not an intervention in a world."""
        self.assertTrue(node_not_an_intervention_in_world(world=frozenset([-x]), node=Y))
        self.assertFalse(node_not_an_intervention_in_world(world=frozenset([-x, +z]), node=X))
        self.assertTrue(node_not_an_intervention_in_world(world=frozenset([-x, +x]), node=Y))
        self.assertFalse(node_not_an_intervention_in_world(world=frozenset([-x]), node=X))
        self.assertFalse(node_not_an_intervention_in_world(world=frozenset([-x, -y]), node=Y))
        self.assertFalse(node_not_an_intervention_in_world(world=frozenset([-x, +y]), node=Y))
        self.assertFalse(node_not_an_intervention_in_world(world=frozenset([-x, +y, -y]), node=Y))
        with self.assertRaises(TypeError):
            node_not_an_intervention_in_world(world=frozenset([-x]), node=+Y)

        with self.assertRaises(TypeError):
            node_not_an_intervention_in_world(world=frozenset({-x}), node=X @ +x)

    def test_stitch_factual_and_dopplegangers(self):
        """Test that factual variables and their dopplegangers are stitched together unless it is intervened upon."""
        self.assert_uedge_set_equal(
            {(Y, Y @ -x)},
            stitch_factual_and_dopplegangers(
                graph=NxMixedGraph.from_edges(directed=[(X, Y)]), worlds=set([frozenset([-x])])
            ),
        )
        self.assert_uedge_set_equal(
            {(Y, Y @ +x)},
            stitch_factual_and_dopplegangers(
                graph=NxMixedGraph.from_edges(directed=[(X, Y)]), worlds=set([frozenset([+x])])
            ),
        )
        self.assert_uedge_set_equal(
            {(Y, Y @ +x), (Z, Z @ +x)},
            stitch_factual_and_dopplegangers(
                graph=NxMixedGraph.from_edges(directed=[(X, Y), (Y, Z)]),
                worlds=set([frozenset([+x])]),
            ),
        )
        self.assert_uedge_set_equal(
            {(Y, Y @ +x), (Z, Z @ +x)},
            stitch_factual_and_dopplegangers(
                graph=NxMixedGraph.from_edges(directed=[(X, Y), (Y, Z)], undirected=[(X, Z)]),
                worlds=set([frozenset([+x])]),
            ),
        )
        self.assert_uedge_set_equal(
            {(Y, Y @ +x), (Z, Z @ +x), (D, D @ +x), (X, X @ -d), (Y, Y @ -d), (Z, Z @ -d)},
            stitch_factual_and_dopplegangers(
                graph=NxMixedGraph.from_edges(
                    directed=[(X, Y), (Y, Z), (D, X)], undirected=[(X, Z)]
                ),
                worlds=set([frozenset([+x]), frozenset([-d])]),
            ),
        )
        self.assert_uedge_set_equal(
            {
                (Y, Y @ +x),
                (Z, Z @ +x),
                (D, D @ +x),
                (X, X @ -d),
                (Y, Y @ -d),
                (Z, Z @ -d),
                (W, W @ -d),
                (W, W @ +x),
            },
            stitch_factual_and_dopplegangers(
                graph=figure_9a.graph, worlds=set([frozenset([+x]), frozenset([-d])])
            ),
        )

    def test_is_not_self_intervened(self):
        """Test that we can detect when a counterfactual variable intervenes on itself."""
        self.assertFalse(is_not_self_intervened(Y @ (+x, -y)))
        self.assertTrue(is_not_self_intervened(Y @ (+x, -z)))
        self.assertFalse(is_not_self_intervened(Y @ (+x, +y)))

    def test_stitch_factual_and_doppleganger_neighbors(self):
        """Test that factual variables and their dopplegangers are stitched together unless it is intervened upon."""
        self.assert_uedge_set_equal(
            set(),
            stitch_factual_and_doppleganger_neighbors(
                graph=NxMixedGraph.from_edges(directed=[(X, Y)]), worlds=set([frozenset([-x])])
            ),
        )
        self.assert_uedge_set_equal(
            {(X, Y @ +x)},
            stitch_factual_and_doppleganger_neighbors(
                graph=NxMixedGraph.from_edges(directed=[(X, Y)], undirected=[(X, Y)]),
                worlds=set([frozenset([+x])]),
            ),
        )
        self.assert_uedge_set_equal(
            {(X, Z @ +x)},
            stitch_factual_and_doppleganger_neighbors(
                graph=NxMixedGraph.from_edges(directed=[(X, Y), (Y, Z)], undirected=[(X, Z)]),
                worlds=set([frozenset([+x])]),
            ),
        )
        self.assert_uedge_set_equal(
            {(X, Z @ +x), (X, Z @ -d), (Z, X @ -d)},
            stitch_factual_and_doppleganger_neighbors(
                graph=NxMixedGraph.from_edges(
                    directed=[(X, Y), (Y, Z), (D, X)], undirected=[(X, Z)]
                ),
                worlds=set([frozenset([+x]), frozenset([-d])]),
            ),
        )

    def test_stitch_counterfactual_and_dopplegangers(self):
        """Test counterfactual variables and their dopplegangers are stitched together unless it is intervened upon."""
        self.assert_uedge_set_equal(
            set(),
            stitch_counterfactual_and_dopplegangers(
                graph=NxMixedGraph.from_edges(directed=[(X, Y)]), worlds=set([frozenset([-x])])
            ),
        )
        self.assert_uedge_set_equal(
            set(),
            stitch_counterfactual_and_dopplegangers(
                graph=NxMixedGraph.from_edges(directed=[(X, Y)]), worlds=set([frozenset([+x])])
            ),
        )
        self.assert_uedge_set_equal(
            set(),
            stitch_counterfactual_and_dopplegangers(
                graph=NxMixedGraph.from_edges(directed=[(X, Y), (Y, Z)]),
                worlds=set([frozenset([+x])]),
            ),
        )
        self.assert_uedge_set_equal(
            set(),
            stitch_counterfactual_and_dopplegangers(
                graph=NxMixedGraph.from_edges(directed=[(X, Y), (Y, Z)], undirected=[(X, Z)]),
                worlds=set([frozenset([+x])]),
            ),
        )
        self.assert_uedge_set_equal(
            {frozenset({Z @ -d, Z @ +x}), frozenset({Y @ +x, Y @ -d})},
            set(
                stitch_counterfactual_and_dopplegangers(
                    graph=NxMixedGraph.from_edges(
                        directed=[(X, Y), (Y, Z), (D, X)], undirected=[(X, Z)]
                    ),
                    worlds=set([frozenset([+x]), frozenset([-d])]),
                )
            ),
        )
        self.assert_uedge_set_equal(
            {frozenset({Y @ -d, Y @ +x}), frozenset({W @ +x, W @ -d}), frozenset({Z @ +x, Z @ -d})},
            stitch_counterfactual_and_dopplegangers(
                graph=figure_9a.graph, worlds=set([frozenset([+x]), frozenset([-d])])
            ),
        )

    def test_stitch_counterfactual_and_doppleganger_neighbors(self):
        """Test that counterfactual variables and their neighbor dopplegangers are stitched together."""
        self.assert_uedge_set_equal(
            set(),
            set(
                stitch_counterfactual_and_doppleganger_neighbors(
                    graph=NxMixedGraph.from_edges(directed=[(X, Y)]), worlds=set([frozenset([-x])])
                )
            ),
        )
        self.assert_uedge_set_equal(
            set(),
            set(
                stitch_counterfactual_and_doppleganger_neighbors(
                    graph=NxMixedGraph.from_edges(directed=[(X, Y)]), worlds=set([frozenset([+x])])
                )
            ),
        )
        self.assert_uedge_set_equal(
            set(),
            set(
                stitch_counterfactual_and_doppleganger_neighbors(
                    graph=NxMixedGraph.from_edges(directed=[(X, Y), (Y, Z)]),
                    worlds=set([frozenset([+x])]),
                )
            ),
        )
        self.assert_uedge_set_equal(
            set(),
            set(
                stitch_counterfactual_and_doppleganger_neighbors(
                    graph=NxMixedGraph.from_edges(directed=[(X, Y), (Y, Z)], undirected=[(X, Z)]),
                    worlds=set([frozenset([+x])]),
                )
            ),
        )
        self.assert_uedge_set_equal(
            {frozenset({X @ -d, Z @ +x})},
            set(
                stitch_counterfactual_and_doppleganger_neighbors(
                    graph=NxMixedGraph.from_edges(
                        directed=[(X, Y), (Y, Z), (D, X)], undirected=[(X, Z)]
                    ),
                    worlds=set([frozenset([+x]), frozenset([-d])]),
                )
            ),
        )
        self.assert_uedge_set_equal(
            {frozenset({X @ -d, Z @ +x}), frozenset({Y @ +x, Z @ -d}), frozenset({Y @ -d, Z @ +x})},
            set(
                stitch_counterfactual_and_doppleganger_neighbors(
                    graph=NxMixedGraph.from_edges(
                        directed=[(X, Y), (Y, Z), (D, X)], undirected=[(X, Z), (Y, Z)]
                    ),
                    worlds=set([frozenset([+x]), frozenset([-d])]),
                )
            ),
        )
        self.assert_uedge_set_equal(
            {frozenset({X @ -d, Y @ +x})},
            stitch_counterfactual_and_doppleganger_neighbors(
                graph=figure_9a.graph, worlds=set([frozenset([+x]), frozenset([-d])])
            ),
        )

    def test_stitch_counterfactual_and_neighbors(self):
        """Test counterfactual variables and their neighbors are stitched together."""
        self.assert_uedge_set_equal(
            set(),
            set(
                stitch_counterfactual_and_neighbors(
                    graph=NxMixedGraph.from_edges(directed=[(X, Y)]), worlds=set([frozenset([-x])])
                )
            ),
        )
        self.assert_uedge_set_equal(
            set(),
            set(
                stitch_counterfactual_and_neighbors(
                    graph=NxMixedGraph.from_edges(directed=[(X, Y)]), worlds=set([frozenset([+x])])
                )
            ),
        )
        self.assert_uedge_set_equal(
            set(),
            set(
                stitch_counterfactual_and_neighbors(
                    graph=NxMixedGraph.from_edges(directed=[(X, Y), (Y, Z)]),
                    worlds=set([frozenset([+x])]),
                )
            ),
        )
        self.assert_uedge_set_equal(
            {frozenset({Y @ +x, Z @ +x})},
            set(
                stitch_counterfactual_and_neighbors(
                    graph=NxMixedGraph.from_edges(
                        directed=[(X, Y), (Y, Z)], undirected=[(X, Z), (Y, Z)]
                    ),
                    worlds=set([frozenset([+x])]),
                )
            ),
        )
        self.assert_uedge_set_equal(
            {frozenset({X @ -d, Z @ -d})},
            set(
                stitch_counterfactual_and_neighbors(
                    graph=NxMixedGraph.from_edges(
                        directed=[(X, Y), (Y, Z), (D, X)], undirected=[(X, Z)]
                    ),
                    worlds=set([frozenset([+x]), frozenset([-d])]),
                )
            ),
        )
        self.assert_uedge_set_equal(
            {frozenset({Y @ +x, Z @ +x}), frozenset({Y @ -d, Z @ -d}), frozenset({Z @ -d, X @ -d})},
            set(
                stitch_counterfactual_and_neighbors(
                    graph=NxMixedGraph.from_edges(
                        directed=[(X, Y), (Y, Z), (D, X)], undirected=[(X, Z), (Y, Z)]
                    ),
                    worlds=set([frozenset([+x]), frozenset([-d])]),
                )
            ),
        )
        self.assert_uedge_set_equal(
            {frozenset({X @ -d, Y @ -d})},
            stitch_counterfactual_and_neighbors(
                graph=figure_9a.graph, worlds=set([frozenset([+x]), frozenset([-d])])
            ),
        )

    def test_get_directed_edges(self):
        """Test that the directed edges of a parallel world graph are correctly identified."""
        self.assertEqual(
            {
                (X @ -d, Y @ -d),
                (Y @ -d, Z @ -d),
                (D @ -d, X @ -d),
                (X @ +x, Y @ +x),
                (Y @ +x, Z @ +x),
            },
            _get_directed_edges(
                NxMixedGraph.from_edges(directed=[(X, Y), (Y, Z), (D, X)]),
                worlds=set([frozenset([+x]), frozenset([-d])]),
            ),
        )

    def test_is_pw_equivalent(self):
        """Test that two nodes in a parallel world graph are the same (lemma 24)."""
        event: Event = {Y @ -x: -y, D: -d, Z @ -d: -z, X: +x}
        self.assertTrue(is_pw_equivalent(figure_9b.graph, event, D @ -X, D))
        self.assertTrue(is_pw_equivalent(figure_9b.graph, event, X @ -D, X))
        self.assertTrue(is_pw_equivalent(figure_11a.graph, event, Z, Z @ -X))
        self.assertTrue(is_pw_equivalent(figure_11a.graph, event, W, W @ -d))
        self.assertTrue(is_pw_equivalent(figure_11a.graph, event, Z @ -x, Z))
        self.assertTrue(is_pw_equivalent(figure_11a.graph, event, Z @ -d, Z))
        self.assertTrue(is_pw_equivalent(figure_11a.graph, event, Z @ -x, Z @ -d))
        self.assertFalse(is_pw_equivalent(figure_11a.graph, event, D @ -d, D))
        self.assertTrue(is_pw_equivalent(figure_9b.graph, event, D @ -x, D))
        self.assertTrue(is_pw_equivalent(figure_11a.graph, event, Z @ -d, Z))
        self.assertFalse(is_pw_equivalent(figure_9b.graph, event, X, X @ -X))
        self.assertFalse(is_pw_equivalent(figure_9b.graph, event, Z, Z @ -X))
        self.assertFalse(is_pw_equivalent(figure_9b.graph, event, D, D @ -d))
        self.assertFalse(is_pw_equivalent(figure_9b.graph, event, X, X @ -x))
        self.assertFalse(is_pw_equivalent(figure_11a.graph, event, X @ -x, X))
        self.assertFalse(
            is_pw_equivalent(
                NxMixedGraph.from_edges(
                    directed=[(X, Y), (X @ +x, Y @ +y), (X @ -x, Y @ -y)],
                    undirected=[(Y @ +y, Y @ -y), (Y @ -y, Y)],
                ),
                {Y @ +x: +y, Y @ -x: -y},
                X @ +x,
                X @ -x,
            )
        )
        self.assertFalse(
            is_pw_equivalent(
                NxMixedGraph.from_edges(
                    directed=[(X, Y), (X @ +x, Y @ +x), (X @ -x, Y @ -x)],
                    undirected=[(Y @ +x, Y @ -x), (Y @ -x, Y)],
                ),
                {Y @ +x: +y, Y @ -x: -y},
                X,
                X @ -x,
            )
        )
        self.assertFalse(
            is_pw_equivalent(
                NxMixedGraph.from_edges(
                    directed=[(X, Y), (X @ +x, Y @ +y), (X @ -x, Y @ -y)],
                    undirected=[(Y @ +y, Y @ -y), (Y @ -y, Y)],
                ),
                {Y @ +x: +y, Y @ -x: -y},
                X,
                X @ +x,
            )
        )
        self.assertFalse(
            is_pw_equivalent(
                NxMixedGraph.from_edges(
                    directed=[(X, Y), (X @ +x, Y @ +x), (X @ -x, Y @ -x)],
                    undirected=[(Y @ +y, Y @ -x), (Y @ -x, Y)],
                ),
                {Y @ +x: +y, Y @ -x: -y},
                Y @ +x,
                Y @ -x,
            )
        )
        self.assertFalse(
            is_pw_equivalent(
                NxMixedGraph.from_edges(
                    directed=[(X, Y), (X @ +x, Y @ +x), (X @ -x, Y @ -x)],
                    undirected=[(Y @ +x, Y @ -x), (Y @ -x, Y)],
                ),
                {Y @ +x: +y, Y @ -x: -y},
                Y @ +x,
                Y,
            )
        )
        self.assertFalse(
            is_pw_equivalent(
                NxMixedGraph.from_edges(
                    directed=[(X, Y), (X @ +x, Y @ +x), (X @ -x, Y @ -x)],
                    undirected=[(Y @ +x, Y @ -x), (Y @ -x, Y)],
                ),
                {Y @ +x: +y, Y @ -x: -y},
                Y,
                Y @ -x,
            )
        )
        self.assertTrue(
            is_pw_equivalent(
                NxMixedGraph.from_edges(
                    directed=[(X, Y), (X @ +x, Y @ +x), (X @ -x, Y @ -x)],
                    undirected=[(Y @ +x, Y @ -x), (Y @ -x, Y)],
                ),
                {Y @ +x: +y, Y @ -x: -y, X: -x},
                Y,
                Y @ -x,
            )
        )

    def test_merge_pw(self):
        """Test the parallel worlds graph after merging two nodes is correct (Lemma 25)."""
        cf_graph_1, preferred, eliminated = merge_pw(figure_9b.graph, D, D @ -X)
        cf_graph_2, preferred, eliminated = merge_pw(cf_graph_1, X, X @ D)
        # test that we swap the order of the nodes if the first is a counterfactual
        cf_graph_3, preferred, eliminated = merge_pw(cf_graph_2, Z @ -X, Z)
        cf_graph_4, preferred, eliminated = merge_pw(cf_graph_3, Z, Z @ D)
        cf_graph_5, preferred, eliminated = merge_pw(cf_graph_4, W, W @ D)
        cf_graph_6, preferred, eliminated = merge_pw(cf_graph_5, D @ -D, D)
        cf_graph_7, preferred, eliminated = merge_pw(cf_graph_6, Y, Y @ -D)
        # test that we sort the order of the nodes if both are counterfactual
        cf_graph_8, preferred8, eliminated8 = merge_pw(figure_9b.graph, W @ -d, W @ -x)
        # test that we sort the order of the nodes if the both are factual
        cf_graph_9, preferred9, eliminated9 = merge_pw(figure_9b.graph, W, Z)
        self.assert_graph_equal(figure_11a.graph, cf_graph_2)
        self.assert_graph_equal(figure_11b.graph, cf_graph_6)
        self.assert_graph_equal(figure_11c.graph, cf_graph_7)
        self.assert_graph_equal(merge_pw(figure_9b.graph, W @ -x, W @ -d)[0], cf_graph_8)
        self.assert_graph_equal(merge_pw(figure_9b.graph, Z, W)[0], cf_graph_9)
        self.assertTrue(D @ -d not in merge_pw(figure_11a.graph, Z, Z @ -d)[0].nodes())

    def test_is_inconsistent(self):
        r"""Test whether two nodes are inconsistent."""
        self.assertTrue(is_inconsistent({D: -d, D @ +x: +d}, D, D @ ~X))
        self.assertTrue(is_inconsistent({D @ -x: -d, D @ +x: +d}, D @ +x, D @ -x))
        self.assertTrue(is_inconsistent({Y @ -x: -y, Y @ +x: +y}, Y @ +x, Y @ -x))

    def test_make_counterfactual_graph(self):
        r"""Test making a counterfactual graph.

        The invocation of **make-cg** with the graph in Figure 9(a) and the joint distribution
        :math:`P(y_x, x', z, d)` will result in the counterfactual graph shown in Fig. 9(c).
        The invocation of **make-cg** with the graph in Figure 9(a) and the joint distribution
        :math:`P(y_{x,z},x')` will result in the counterfactual graph shown in Fig. 9(d).
        """
        actual_graph, actual_event = make_counterfactual_graph(
            figure_9a.graph, {Y @ -x: -y, X: +x, Z @ -d: -z, D: -d}
        )
        self.assert_graph_equal(figure_9c.graph, actual_graph)
        self.assertEqual({Y @ -x: -y, X: +x, Z: -z, D: -d}, actual_event)
        actual_graph2, actual_event2 = make_counterfactual_graph(
            figure_9a.graph, {Y @ (-x, -z): -y, X: +x}
        )
        expected_graph2, expected_event2 = figure_9d.graph, {Y @ (-x, -z): -y, X: +x}
        self.assertEqual(expected_event2, actual_event2)
        self.assert_graph_equal(expected_graph2, actual_graph2)

        # Check for inconsistent counterfactual values for merged nodes
        actual_graph3, actual_event3 = make_counterfactual_graph(
            graph=NxMixedGraph.from_edges(directed=[(D, Z), (Z, Y)]),
            event={Z @ -d: -z, Z: +z, D: -d},
        )
        self.assertIsNone(actual_event3)

        # # Check whether {Y_{+x,z,w): -y, X_w: -x} automatically simplifies to {Y_{z,w}: y, X: -x} (it should not)
        # actual_graph4, actual_event4 = make_counterfactual_graph(
        #     graph=figure_9a.graph, event={X @ -W: X, Y @ (-W, +X, -Z): Y}
        # )
        # expected_event4 = {Y @ (-W, -Z): Y, X: X}
        # expected_graph4 = NxMixedGraph.from_edges(
        #     nodes={W @ -W, Y @ (-W, +X, -Z), Z @ (-W, +X, -Z), X},
        #     directed={(W @ -W, Y @ (-W, +X, -Z)), (Z @ (-W, +X, -Z), Y @ (-W, +X, -Z))},
        #     undirected={frozenset({X, Y @ (-W, +X, -Z)})},
        # )
        # self.assertNotEqual(expected_event4, actual_event4)
        # self.assert_graph_equal(expected_graph4, actual_graph4)

        # Check whether the counterfactual graph is consistent (it is not)
        actual_graph5, actual_event5 = make_counterfactual_graph(
            graph=NxMixedGraph.from_edges(directed=[(W, X), (Z, X), (X, Y)]),
            event={Y @ -w: -y, Y @ -z: +y, X @ -w: +x, X @ -z: +x},
        )
        self.assertIsNone(actual_event5)

        # Check whether Probability of necessary and sufficient causation induces a W graph
        actual_graph6, actual_event6 = make_counterfactual_graph(
            graph=NxMixedGraph.from_edges(directed=[(X, Y)]), event={Y @ -x: -y, Y @ +x: +y}
        )
        expected_graph6 = NxMixedGraph.from_edges(
            directed=[(X @ -x, Y @ -x), (X @ +x, Y @ +x)], undirected=[(Y @ -x, Y @ +x)]
        )

        expected_event6 = {Y @ -x: -y, Y @ +x: +y}
        self.assert_graph_equal(expected_graph6, actual_graph6)
        self.assertEqual(expected_event6, actual_event6)

        # Check that a triplet world graph is not inconsistent
        actual_graph7, actual_event7 = make_counterfactual_graph(
            graph=NxMixedGraph.from_edges(
                directed=[(X, W), (W, Y), (D, Z), (Z, Y), (A, B), (B, Y)], undirected=[(X, Y)]
            ),
            event={Y @ +x: +y, X: -x, Z @ -d: -z, D: -d, A: +A},
        )
        self.assertIsNotNone(actual_event7)
        expected_graph7 = NxMixedGraph.from_edges(
            directed=[(D, Z), (B, Y @ +X), (W @ +X, Y @ +X), (Z, Y @ +X), (A, B), (X @ +X, W @ +X)],
            undirected=[(X, Y @ +X)],
        )
        self.assert_graph_equal(expected_graph7, actual_graph7)

        actual_graph8, actual_event8 = make_counterfactual_graph(
            graph=NxMixedGraph.from_edges(directed=[(X, Z), (Z, Y)]),
            event={Y @ -x: -y, Y @ +x: -y, Z @ +x: -z, Z @ -x: -z},
        )
        expected_graph8 = NxMixedGraph.from_edges(
            directed=[(X @ -x, Z @ -x), (Z @ -x, Y @ -x), (X @ +x, Z @ +x)],
            undirected=[(Z @ -x, Z @ +x)],
        )

        expected_event8 = {Y @ -X: -Y, Z @ +X: -Z, Z @ -X: -Z}
        self.assert_graph_equal(expected_graph8, actual_graph8)
        self.assertEqual(expected_event8, actual_event8)



# -*- coding: utf-8 -*-

"""Test cases."""

import unittest

from y0.dsl import Expression, get_outcomes_and_treatments
from y0.graph import NxMixedGraph
from y0.mutate import canonicalize

__all__ = ["GraphTestCase"]


class GraphTestCase(unittest.TestCase):
    """Tests parallel worlds and counterfactual graphs."""

    def assert_graph_equal(self, expected: NxMixedGraph, actual: NxMixedGraph, msg=None) -> None:
        """Check the graphs are equal (more nice than the builtin :meth:`NxMixedGraph.__eq__` for testing)."""
        self.assertEqual(set(expected.directed.nodes()), set(actual.directed.nodes()), msg=msg)
        self.assertEqual(set(expected.undirected.nodes()), set(actual.undirected.nodes()), msg=msg)
        self.assertEqual(set(expected.directed.edges()), set(actual.directed.edges()), msg=msg)
        self.assertEqual(
            set(map(frozenset, expected.undirected.edges())),
            set(map(frozenset, actual.undirected.edges())),
            msg=msg,
        )

    def assert_expr_equal(self, expected: Expression, actual: Expression) -> None:
        """Assert that two expressions are the same."""
        expected_outcomes, expected_treatments = get_outcomes_and_treatments(query=expected)
        actual_outcomes, actual_treatments = get_outcomes_and_treatments(query=actual)
        self.assertEqual(expected_treatments, actual_treatments)
        self.assertEqual(expected_outcomes, actual_outcomes)
        ordering = sorted(expected.get_variables(), key=lambda x: str(x))
        expected_canonical = canonicalize(expected, ordering)
        actual_canonical = canonicalize(actual, ordering)
        self.assertEqual(
            expected_canonical,
            actual_canonical,
            msg=f"\nExpected: {str(expected_canonical)}\nActual:   {str(actual_canonical)}",
        )



# -*- coding: utf-8 -*-

"""Calculate the complexity of an expression."""

from typing import Iterable

from .dsl import (
    CounterfactualVariable,
    Expression,
    Fraction,
    Intervention,
    One,
    Probability,
    Product,
    QFactor,
    Sum,
    Variable,
    Zero,
)

__all__ = [
    "complexity",
]

CONST_CONST = 1.0  # yo dawg
VAR_CONSTANT = 1.0  # yo cat
SUM_CONST = 0.0
FRAC_CONST = 0.0
PROD_CONST = 0.0
PROB_CONST = 0.0
Q_CONST = 0.0


def complexity(expr: Expression) -> float:
    """Calculate the "complexity" of the expression, where a bigger result means more complex.

    Currently, the complexity of the expression is defined recursively, where all of the
    parts of each expression are added up, counting the number of variables in each.

    :param expr: The input expression
    :return: The complexity of the expression
    :raises TypeError:
        Raised if an invalid expression type is used
    """
    if isinstance(expr, (One, Zero)):
        return CONST_CONST
    if isinstance(expr, Fraction):
        return FRAC_CONST + complexity(expr.numerator) + complexity(expr.denominator)
    if isinstance(expr, Product):
        return PROD_CONST + sum(complexity(subexpr) for subexpr in expr.expressions)
    if isinstance(expr, Sum):
        return SUM_CONST + range_complexity(expr.ranges) + complexity(expr.expression)
    if isinstance(expr, QFactor):
        return Q_CONST + range_complexity(expr.domain) + range_complexity(expr.codomain)
    if isinstance(expr, Probability):
        return PROB_CONST + probability_complexity(expr)
    raise TypeError(f"Unhandled expression type: {expr.__class__.__name__}")


def range_complexity(variables: Iterable[Variable]) -> float:
    """Return the complexity of a sequence of variables."""
    return sum(variable_complexity(v) for v in variables)


def variable_complexity(variable: Variable) -> float:
    """Return the complexity of a single variable."""
    if isinstance(variable, CounterfactualVariable):
        return VAR_CONSTANT + sum(variable_complexity(v) for v in variable.interventions)
    elif isinstance(variable, Intervention):
        return VAR_CONSTANT
    else:
        return VAR_CONSTANT


def probability_complexity(expr: Probability) -> float:
    """Return the complexity of a probability expression."""
    return range_complexity(expr.children) + range_complexity(expr.parents)



"""This file contains code that implements the original ID* algorithm.

The algorithm is described in the paper "Complete Identification Methods for the Causal Hierarchy"
but the algorithm itself is not complete, as there are identifiable queries that cannot be
identified with this algorithm.
"""

import logging
from typing import Collection, Set, Tuple, cast

from y0.algorithm.identify.cg import is_not_self_intervened, make_counterfactual_graph
from y0.algorithm.identify.id_star import (
    ConflictUnidentifiable,
    District,
    DistrictInterventions,
    get_conflicts,
    get_free_variables,
    id_star_line_9,
    remove_event_tautologies,
    violates_axiom_of_effectiveness,
)
from y0.dsl import Event, Expression, Intervention, One, Product, Sum, Variable, Zero
from y0.graph import NxMixedGraph

logger = logging.getLogger(__name__)


def original_id_star_line_6(
    graph: NxMixedGraph, event: Event
) -> Tuple[Collection[Variable], DistrictInterventions]:
    r"""Run line 6 of the ID* algorithm.

    Line 6 is analogous to Line 4 in the ID algorithm, it decomposes
    the problem into a set of subproblems, one for each C-component in
    the counterfactual graph. In the ID algorithm, the term
    corresponding to a given C-component :math:`S_i` of the causal
    diagram was the effect of all variables not in :math:`S_i` on
    variables in :math:`S_i` , in other words
    :math:`P_{\mathbf{v}\backslash s_i} (s_i )`, and the outermost
    summation on line 4 was over values of variables not in
    :math:`\mathbf{Y},\mathbf{X}`. Here, the term corresponding to a
    given C-component :math:`S^i` of the counterfactual graph
    :math:`G'` is the conjunction of counterfactual variables where
    each variable contains in its subscript all variables not in the
    C-component :math:`S^i` , in other words :math:`\mathbf{v}(G'
    )\backslash s^i` , and the outermost summation is over observable
    variables not in :math:`\event'` , that is over
    :math:`\mathbf{v}(G' ) \backslash \event'` , where we interpret
    :math:`\event'` as a set of counterfactuals, rather than a
    conjunction.

    Unfortunately, the conjunction of all counterfactual variables in
    the C-component where each variable contains its subscript over
    all variables not in the C-component results in a problem.  The
    problem is that the query contains redundant interventions that
    must be removed for the query to be identifiable.  The removal of
    redundant interventions is not specified as part of the algorithm
    itself.  Instead, it must be specified outside the algorithm.  The
    algorithm is not complete without this step.  Tikka et al. (2022)
    address this problem by removing the rudundant interventions in
    step 7 of the algorithm after running ID* on each C-component.
    However, it is simpler to not include redundant interventions in
    the first place.  Instead of intervening on every variable not in
    the C-component, we intervene on every variable that is a parent
    of the C-component.  This is also known as the Markov pillow. Zucker
    suggested this approach in an email to Shpitser on May 20, 2022. In
    this function we demonstrate what happens when we intervene on
    every variable not in the C-component.

    :param graph: an NxMixedGraph
    :param event: a conjunction of counterfactual variables
    :return: a set of Variables in summand, a dictionary of districts and events
    """
    # First we get the summand, then we intervene on each district
    summand = get_free_variables(graph, event)
    interventions_of_each_district = get_district_interventions(graph, event)
    return summand, interventions_of_each_district


def get_district_interventions(graph: NxMixedGraph, event: Event) -> DistrictInterventions:
    """For each district, intervene on the variables not in the district.

    Self-interventions are not considered part of the district

    :param graph: an NxMixedGraph
    :param event: a conjunction of counterfactual variables
    :return: a dictionary of districts and interventions of districts

    """
    nodes = set(node for node in graph.nodes() if is_not_self_intervened(node))
    return {
        district: intervene_on_district(district, nodes - district, event)
        for district in graph.subgraph(nodes).districts()
    }


def intervene_on_district(district: District, interventions: Set[Variable], event: Event) -> Event:
    """Intervene on the variables not in the district.

    Each variable in the district takes its value in the event if it has one.
    Otherwise, we set the value to ``-variable.get_base()``.

    :param district: a district of the graph
    :param event: a conjunction of counterfactual values
    :param interventions: a set of Variables to act as interventions
    :returns: A new event
    """
    interventions = {-i.get_base() for i in interventions}
    return {
        variable.intervene(interventions): (
            event[variable] if variable in event else cast(Intervention, -variable.get_base())
        )
        for variable in district
    }


def original_id_star(
    graph: NxMixedGraph, event: Event, *, _number_recursions: int = 0
) -> Expression:
    """Apply the ``ID*`` algorithm to the graph."""
    logger.debug(
        "[%d]: Calling ID* algorithm with graph G with\n\t nodes: %s\n"
        "\t directed: %s\n\t undirected %s\n"
        "\t outcome event: %s",
        _number_recursions,
        graph.nodes(),
        graph.directed.edges(),
        graph.undirected.edges(),
        event,
    )
    # Line 1: There's nothing in the counterfactual event (i.e., an empty conjunction),
    # then its probability is 1, by convention.
    if not event:
        return One()
    # Line 2: This violates the Axiom of Effectiveness
    if violates_axiom_of_effectiveness(event):
        return Zero()
    # Line 3: This is a tautological event and can be removed without affecting the probability
    reduced_event = remove_event_tautologies(event)
    if reduced_event != event:
        logger.debug("[%d] recurring on reduced event %s", _number_recursions, reduced_event)
        return original_id_star(graph, reduced_event, _number_recursions=_number_recursions + 1)
    # Line 4: invokes make-cg to construct a counterfactual graph :math:`G'` , and the
    # corresponding relabeled counterfactual event.
    cf_graph, new_event = make_counterfactual_graph(graph, event)
    logger.debug(
        "[%d] ID* Returned from make_counterfactual_graph(). New event: %s\n"
        "\tcounterfactual graph:\n"
        "\t nodes: %s\n"
        "\t directed: %s\n"
        "\t undirected: %s",
        _number_recursions,
        new_event,
        cf_graph.nodes(),
        cf_graph.directed.edges(),
        cf_graph.undirected.edges(),
    )
    # Line 5:
    if new_event is None:
        return Zero()

    # Line 6:
    nodes = set(node for node in cf_graph.nodes() if is_not_self_intervened(node))
    cf_subgraph = cf_graph.subgraph(nodes)
    if not cf_subgraph.is_connected():
        summand, events_of_each_district = original_id_star_line_6(cf_graph, new_event)
        logger.debug("[%d] summand: %s", _number_recursions, summand)
        assert 1 < len(events_of_each_district)
        logger.debug(
            "[%d] recurring on each district: %s ", _number_recursions, events_of_each_district
        )
        return Sum.safe(
            Product.safe(
                original_id_star(
                    graph, events_of_district, _number_recursions=_number_recursions + 1
                )
                for events_of_district in events_of_each_district.values()
            ),
            summand,
        )

    # Line 7:
    conflicts = get_conflicts(cf_subgraph, new_event)
    if conflicts:
        raise ConflictUnidentifiable(cf_subgraph, new_event, conflicts)

    # Line 9
    return id_star_line_9(cf_subgraph)



# -*- coding: utf-8 -*-

"""An implementation of the identification algorithm."""

from typing import List, Sequence, Union

from .utils import Identification, Query, Unidentifiable
from ...dsl import Expression, P, Probability, Product, Sum, Variable
from ...graph import NxMixedGraph, _ensure_set

__all__ = [
    "identify_outcomes",
    "identify",
]


def identify_outcomes(
    graph: NxMixedGraph,
    treatments: Union[Variable, set[Variable]],
    outcomes: Union[Variable, set[Variable]],
) -> Expression | None:
    """Get the estimand for the outcomes given the treatments."""
    treatments = _ensure_set(treatments)
    outcomes = _ensure_set(outcomes)
    try:
        rv = identify(
            Identification(graph=graph, query=Query(treatments=treatments, outcomes=outcomes))
        )
    except Unidentifiable:
        return None
    return rv


def identify(identification: Identification) -> Expression:
    """Run the identification algorithm.

    :param identification: The identification tuple
    :returns: the expression corresponding to the identification
    :raises Unidentifiable: If no appropriate identification can be found
    """
    graph = identification.graph
    treatments = identification.treatments
    outcomes = identification.outcomes
    vertices = set(graph.nodes())

    # line 1
    if not treatments:
        return line_1(identification)

    # line 2
    outcomes_and_ancestors = graph.ancestors_inclusive(outcomes)
    not_outcomes_or_ancestors = vertices.difference(outcomes_and_ancestors)
    if not_outcomes_or_ancestors:
        return identify(line_2(identification))

    # line 3
    no_effect_on_outcome = graph.get_no_effect_on_outcomes(treatments, outcomes)
    if no_effect_on_outcome:
        return identify(line_3(identification))

    # line 4
    graph_without_treatments = graph.remove_nodes_from(treatments)
    if not graph_without_treatments.is_connected():
        expression = Product.safe(map(identify, line_4(identification)))
        return Sum.safe(
            expression=expression,
            ranges=vertices.difference(outcomes | treatments),
        )

    # line 5
    if graph.is_connected():  # e.g., there's only 1 c-component, and it encompasses all vertices
        raise Unidentifiable(graph.nodes(), graph_without_treatments.districts())

    # line 6
    district_without_treatment = _get_single_district(graph_without_treatments)

    if district_without_treatment in graph.districts():
        parents = list(graph.topological_sort())
        expression = Product.safe(p_parents(v, parents) for v in district_without_treatment)
        ranges = district_without_treatment - outcomes
        return Sum.safe(
            expression=expression,
            ranges=ranges,
        )

    # line 7
    return identify(line_7(identification))


def _get_single_district(graph: NxMixedGraph) -> frozenset[Variable]:
    districts = graph.districts()
    if len(districts) != 1:
        raise RuntimeError
    return districts.pop()


def line_1(identification: Identification) -> Expression:
    r"""Run line 1 of identification algorithm.

    If no action has been taken, the effect on :math:`\mathbf Y` is just the marginal of
    the observational distribution

    :param identification: The data structure with the treatment, outcomes, estimand, and graph
    :returns:  The marginal of the outcome variables
    """
    outcomes = identification.outcomes
    vertices = set(identification.graph.nodes())
    return Sum.safe(
        expression=identification.estimand,
        ranges=vertices.difference(outcomes),
    )


def line_2(identification: Identification) -> Identification:
    r"""Run line 2 of the identification algorithm.

    If we are interested in the effect on :math:`\mathbf Y`, it is sufficient to restrict our attention
    on the parts of the model ancestral to :math:`\mathbf Y`.

    .. math::

        \text{if }\mathbf V - An(\mathbf Y)_G \neq \emptyset \\
        \text{ return } \mathbf{ ID}\left(\mathbf y, \mathbf x\cap An(\mathbf Y)_G, \sum_{\mathbf V -
         An(Y)_G}P, G_{An(\mathbf Y)}\right)

    :param identification: The data structure with the treatment, outcomes, estimand, and graph
    :returns: The new estimand
    :raises ValueError: If the line 2 precondition is not met
    """
    graph = identification.graph
    treatments = identification.treatments
    outcomes = identification.outcomes

    vertices = set(graph.nodes())
    outcomes_and_ancestors = graph.ancestors_inclusive(outcomes)
    not_outcomes_or_ancestors = vertices.difference(outcomes_and_ancestors)
    outcome_ancestral_graph = graph.subgraph(outcomes_and_ancestors)

    if not not_outcomes_or_ancestors:
        raise ValueError("line 2 precondition not met")
    return Identification.from_parts(
        outcomes=outcomes,
        treatments=treatments & outcomes_and_ancestors,
        estimand=Sum.safe(expression=identification.estimand, ranges=not_outcomes_or_ancestors),
        graph=outcome_ancestral_graph,
    )


def line_3(identification: Identification) -> Identification:
    r"""Run line 3 of the identification algorithm.

    Forces an action on any node where such an action would have no
    effect on :math:\mathbf Y`—assuming we already acted on
    :math:`\mathbf X`. Since actions remove incoming arrows, we can
    view line 3 as simplifying the causal graph we consider by
    removing certain arcs from the graph, without affecting the
    overall answer.

    :param identification: The data structure with the treatment, outcomes, estimand, and graph
    :returns: The new estimand
    :raises ValueError: If the preconditions for line 3 aren't met.
    """
    outcomes = identification.outcomes
    treatments = identification.treatments
    graph = identification.graph

    no_effect_on_outcome = graph.get_no_effect_on_outcomes(treatments, outcomes)
    if not no_effect_on_outcome:
        raise ValueError(
            'Line 3 precondition not met. There were no variables in "no_effect_on_outcome"'
        )

    return identification.with_treatments(no_effect_on_outcome)


def line_4(identification: Identification) -> List[Identification]:
    r"""Run line 4 of the identification algorithm.

    The key line of the algorithm, it decomposes the problem into a set
    of smaller problems using the key property of *c-component
    factorization* of causal models. If the entire graph is a single
    C-component already, further problem decomposition is impossible,
    and we must provide base cases. :math:`\mathbf{ID}` has three base
    cases.

    :param identification: The data structure with the treatment, outcomes, estimand, and graph
    :returns: A list of new estimands
    :raises ValueError: If the precondition that there are more than 1 districts without treatments is not met
    """
    treatments = identification.treatments
    estimand = identification.estimand
    graph = identification.graph
    vertices = set(graph.nodes())

    # line 4
    graph_without_treatments = graph.remove_nodes_from(treatments)
    districts_without_treatment = graph_without_treatments.districts()
    if len(districts_without_treatment) <= 1:
        raise ValueError("Line 4 precondition not met")
    return [
        Identification.from_parts(
            outcomes=set(district_without_treatment),
            treatments=vertices - district_without_treatment,
            estimand=estimand,
            graph=graph,
        )
        for district_without_treatment in districts_without_treatment
    ]


def line_5(identification: Identification) -> None:
    r"""Run line 5 of the identification algorithm.

    Fails because it finds two C-components, the graph :math:`G`
    itself, and a subgraph :math:`S` that does not contain any
    :math:`\mathbf X` nodes. But that is exactly one of the properties
    of C-forests that make up a hedge. In fact, it turns out that it
    is always possible to recover a hedge from these two c-components.

    :param identification: The data structure with the treatment, outcomes, estimand, and graph
    :raises Unidentifiable: If line 5 realizes that identification is not possible
    """
    treatments = identification.treatments
    graph = identification.graph
    vertices = set(graph.nodes())
    graph_without_treatments = graph.remove_nodes_from(treatments)
    districts_without_treatment = graph_without_treatments.districts()

    # line 5
    districts = graph.districts()
    if districts == {frozenset(vertices)}:
        raise Unidentifiable(districts, districts_without_treatment)


def line_6(identification: Identification) -> Expression:
    r"""Run line 6 of the identification algorithm.

    Asserts that if there are no bidirected arcs from :math:`X` to the other nodes in the current subproblem
    under consideration, then we can replace acting on :math:`X` by conditioning, and thus solve the subproblem.

    ..math::

        \text{ if }S\in C(G) \\
        \text{ return }\sum_{S - \mathbf y}\prod_{\{i|V_i\in S\}}P\left(v_i|v_\pi^{(i-1)}\right)

    :param identification: The data structure with the treatment, outcomes, estimand, and graph
    :returns: A list of new estimands
    :raises ValueError: If line 6 precondition is not met
    """
    outcomes = identification.outcomes
    treatments = identification.treatments
    graph = identification.graph

    districts = graph.districts()
    graph_without_treatments = graph.remove_nodes_from(treatments)
    district_without_treatments = _get_single_district(graph_without_treatments)

    # line 6
    if district_without_treatments not in districts:
        raise ValueError("Line 6 precondition not met")

    parents = list(graph.topological_sort())
    expression = Product.safe(p_parents(v, parents) for v in district_without_treatments)
    ranges = district_without_treatments - outcomes
    return Sum.safe(
        expression=expression,
        ranges=ranges,
    )


def line_7(identification: Identification) -> Identification:
    r"""Run line 7 of the identification algorithm.

    The most complex case where :math:`\mathbf X` is partitioned into
    two sets, :math:`\mathbf W` which contain bidirected arcs into
    other nodes in the subproblem, and :math:`\mathbf Z` which do
    not. In this situation, identifying :math:`P(\mathbf y|do(\mathbf
    x))` from :math:`P(v)` is equivalent to identifying
    :math:`P(\mathbf y|do(\mathbf w))` from :math:`P(\mathbf
    V|do(\mathbf z))`, since :math:`P(\mathbf y|do(\mathbf x)) =
    P(\mathbf y|do(\mathbf w), do(\mathbf z))`. But the term
    :math:`P(\mathbf V|do(\mathbf z))` is identifiable using the
    previous base case, so we can consider the subproblem of
    identifying :math:`P(\mathbf y|do(\mathbf w))`

    .. math::

       \text{ if }(\exists S')S\subset S'\in C(G) \\
       \text{ return }\mathbf{ID}\left(\mathbf y, \mathbf x\cap S',
       \prod_{\{i|V_i\in S'\}}P(V_i|V_\pi^{(i-1)}\cap S', V_\pi^{(i-1)} -
       S'), G_{S'}\right)

    :param identification: The data structure with the treatment, outcomes, estimand, and graph
    :returns: A new estimand
    :raises ValueError: If line 7 does not find a suitable district
    """
    outcomes = identification.outcomes
    treatments = identification.treatments
    graph = identification.graph

    graph_without_treatments = graph.remove_nodes_from(treatments)
    # line 7 precondition requires single district
    district_without_treatments = _get_single_district(graph_without_treatments)

    # line 7
    for district in graph.districts():
        if district_without_treatments < district:
            parents = list(graph.topological_sort())
            return Identification.from_parts(
                outcomes=outcomes,
                treatments=treatments & district,
                estimand=Product.safe(p_parents(v, parents) for v in district),
                graph=graph.subgraph(district),
            )

    raise ValueError("Could not identify suitable district")


def p_parents(child: Variable, ordering: Sequence[Variable]) -> Probability:
    """Get a probability expression based on a topological ordering.

    :param child: The child variable
    :param ordering: A topologically ordered sequence of all variables. All occurring before the
        child will be used as parents.
    :return: A probability expression
    """
    return P(child | ordering[: ordering.index(child)])
